/*
 * atanks - obliterate each other with oversize weapons
 * Copyright (C) 2002,2003  Thomas Hudson,Juraj Michalek
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * */


#include "globals.h"

#define SWITCH_HELP		"-h"
#define SWITCH_FULL_SCREEN	"-fs"
#define SWITCH_WINDOWED		"-win"

int screen_mode = GFX_AUTODETECT_WINDOWED;

BITMAP *create_gradient_strip (gradient *gradient, int length);
int draw_circlesBG (GLOBALDATA *global, BITMAP *dest, int x, int y, int width, int height);

using namespace std;

void fpsadd()
{
	fps = frames;
	frames = 0;
}

void clockadd()
{
	cclock++;
}
END_OF_FUNCTION(clockadd);

double colorDistance (int col1, int col2)
{
	double distance;
	int col1r, col1g, col1b;
	int col2r, col2g, col2b;

	col1r = getr (col1);
	col1g = getg (col1);
	col1b = getb (col1);
	col2r = getr (col2);
	col2g = getg (col2);
	col2b = getb (col2);

	// Treat the colour-cube as a space
	distance = vector_length_f ((float)(col1r - col2r), (float)(col1g - col2g), (float)(col1b - col2b));

	return (distance);
}

double Noise (int x)
{
	x = (x<<13) ^ x;
	return ( 1.0 - ((x * (x * x * 15731 + 789221) + 1376312589) & 0x7fffffff) / 1073741824.0);    
}

double Noise2D (int x, int y)
{
    int n;

    n = x + y * 57;
    n = (n << 13) ^ n;
    return ( 1.0 - ( (n * (n * n * 15731 + 789221) + 1376312589) & 0x7fffffff) / 1073741824.0);
}

double interpolate (double x1, double x2, double i)
{
	double ft = i * M_PI;
	double f = (1 - cos(ft)) * 0.5;

	return (x1 * (1 - f) + (x2 * f));
}

/*
* 
* For a really good explanation of perlin noise (and where most of this
*   perlin code was adapted from) see:
* 
* http://freespace.virgin.net/hugo.elias/models/m_perlin.htm
* 
* It must be good because I understood it <:-)
* 
* - Tom Hudson
* 
*/
double perlin2DPoint (double amplitude, double scale, double xo, double yo, double lambda, int octaves)
{
	double	maxH = 0;
	double	h = 0;
	for (int iteration = 1; iteration <= octaves; iteration++) {
		double	zoom = scale / (iteration * iteration);
		double	fractX = xo / zoom;
		double	fractY = yo / zoom;
		double	h1 = Noise2D ((int)fractX, (int)fractY);
		double	h2 = Noise2D ((int)fractX + 1, (int)fractY);
		double	h3 = Noise2D ((int)fractX, (int)fractY + 1);
		double	h4 = Noise2D ((int)fractX + 1, (int)fractY + 1);

		double	xi = fractX - (int)fractX;
		double	yi = fractY - (int)fractY;

		double	i1 = interpolate (h1, h2, xi);
		double	i2 = interpolate (h3, h4, xi);
		double	i3 = interpolate (i1, i2, yi);

		h += amplitude * i3;
		maxH += amplitude;

		amplitude *= lambda;
	}

	// Normalised
	return (h / maxH);
}

double perlin1DPoint (double amplitude, double scale, double xo, double lambda, int octaves)
{
	double	maxH = 0;
	double	h = 0;
	for (int iteration = 1; iteration <= octaves; iteration++) {
		double	zoom = scale / (iteration * iteration);
		double	fractX = xo / zoom;
		double	h1 = Noise ((int)fractX);
		double	h2 = Noise ((int)fractX + 1);
		double	i = fractX - (int)fractX;

		h += amplitude * interpolate (h1, h2, i);
		maxH += amplitude;

		amplitude *= lambda;
	}

	// Normalised
	return (h / maxH);
}

void drawMenuBackground (GLOBALDATA *global, ENVIRONMENT *env, int itemType, int tOffset, int numItems)
{
	rectfill (env->db, HALF_WIDTH - 300, 100,
		HALF_WIDTH + 300, SCREEN_HEIGHT - 100,
		makecol16 (0,79,0));
	rect     (env->db, HALF_WIDTH - 300, 100,
		HALF_WIDTH + 300, SCREEN_HEIGHT - 100,
		makecol16 (128,255,128));
	
	drawing_mode (DRAW_MODE_TRANS, NULL, 0, 0);
	set_trans_blender (0, 0, 0, 15);
	for (int tCount = 0; tCount < numItems; tCount++) {
		int radius, xpos, ypos;
		switch (itemType) {
			case 0: // circles
				radius = (int)((perlin1DPoint (1.0, 5, (tOffset * 0.02) + tCount + 423346, 0.5, 8) + 1) / 2 * 40);
				xpos = HALF_WIDTH + (int)(perlin1DPoint (1.0, 3, (tOffset * 0.02) + tCount + 232662, 0.3, 6) * 250);
				ypos = HALF_HEIGHT + (int)(perlin1DPoint (1.0, 2, (tOffset * 0.02) + tCount + 42397, 0.3, 6) * (HALF_HEIGHT - 100));
				circlefill (env->db, xpos, ypos, radius,
						makecol (200,255,200));
				break;
			case 1: // Horz lines
				radius = (int)((perlin1DPoint (1.0, 5, (tOffset * 0.02) + tCount + 423346, 0.5, 8) + 1) / 2 * 40);
				xpos = HALF_WIDTH + (int)(perlin1DPoint (1.0, 3, (tOffset * 0.02) + tCount + 232662, 0.3, 6) * 250);
				rectfill (env->db, xpos - radius / 2, 101,
					xpos + radius / 2, SCREEN_HEIGHT - 101,
					makecol (200,255,200));
				break;
			default:
				break;
		}

	}
	solid_mode ();

}


void initialisePlayers (GLOBALDATA *global, ENVIRONMENT *env)
{
	for (z = 0; z < global->numPlayers; z++) {
		global->players[z]->money = (int)global->startmoney;
		global->players[z]->score = 0;
		global->players[z]->initialise ();
	}
}

void spiralPixelFade (BITMAP *target) {
	int blockSize = 8;
	int spiralSize = 20;
	int length = spiralSize, x = 0, y = 0;
	int stepSize = (spiralSize + 1) * blockSize;
	int x2, y2;
	int count;

	for (count = 0, y = stepSize - blockSize; count < length; count++, y -= blockSize)
		for (x2 = 0; x2 <= SCREEN_WIDTH; x2 += stepSize)
			for (y2 = 0; y2 <= SCREEN_HEIGHT; y2 += stepSize)
				blit (target, screen, x2+x, y2+y, x2+x, y2+y, blockSize, blockSize);
	while (length) {
		for (count = 0; count < length; count++, x += blockSize)
			for (x2 = 0; x2 <= SCREEN_WIDTH; x2 += stepSize)
				for (y2 = 0; y2 <= SCREEN_HEIGHT; y2 += stepSize)
					blit (target, screen, x2+x, y2+y, x2+x, y2+y, blockSize, blockSize);
		for (count = 0; count < length; count++, y += blockSize)
			for (x2 = 0; x2 <= SCREEN_WIDTH; x2 += stepSize)
				for (y2 = 0; y2 <= SCREEN_HEIGHT; y2 += stepSize)
					blit (target, screen, x2+x, y2+y, x2+x, y2+y, blockSize, blockSize);
		length--;
		for (count = 0; count < length; count++, x -= blockSize)
			for (x2 = 0; x2 <= SCREEN_WIDTH; x2 += stepSize)
				for (y2 = 0; y2 <= SCREEN_HEIGHT; y2 += stepSize)
					blit (target, screen, x2+x, y2+y, x2+x, y2+y, blockSize, blockSize);
		for (count = 0; count < length; count++, y -= blockSize)
			for (x2 = 0; x2 <= SCREEN_WIDTH; x2 += stepSize)
				for (y2 = 0; y2 <= SCREEN_HEIGHT; y2 += stepSize)
					blit (target, screen, x2+x, y2+y, x2+x, y2+y, blockSize, blockSize);
		length--;
	}
	for (x2 = 0; x2 <= SCREEN_WIDTH; x2 += stepSize)
		for (y2 = 0; y2 <= SCREEN_HEIGHT; y2 += stepSize)
			blit (target, screen, x2+x, y2+y, x2+x, y2+y, blockSize, blockSize);
}

void topDownMultiWipeFade (BITMAP *target) {
	for (zz = 0; zz < 8; zz++) {
		int lineCount = 0;
		for (z = zz; z < SCREEN_HEIGHT; z += 8) {
			blit (target, screen, 0, z, 0, z, SCREEN_WIDTH, 1);
			if (lineCount == 10) {
				lineCount = 0;
				rest (1);
			}
			lineCount++;
		}
	}
}

void topDownWipeFade (BITMAP *target) {
	for (z = 0; z <= SCREEN_HEIGHT / 8; z++) {
		blit (target, screen, 0, z * 8, 0, z * 8, SCREEN_WIDTH, 8);
		rest (1);
	}
}

void bottomUpWipeFade (BITMAP *target) {
	for (z = SCREEN_HEIGHT / 8; z >= 0; z--) {
		blit (target, screen, 0, z * 8, 0, z * 8, SCREEN_WIDTH, 8);
		rest (1);
	}
}

void rightLeftWipeFade (BITMAP *target) {
	for (z = SCREEN_WIDTH / 8; z >= 0; z--) {
		blit (target, screen, z * 8, 0, z * 8, 0, 8, SCREEN_HEIGHT);
		rest (1);
	}
}

void leftRightWipeFade (BITMAP *target) {
	for (z = 0; z <= SCREEN_WIDTH / 8; z++) {
		blit (target, screen, z * 8, 0, z * 8, 0, 8, SCREEN_HEIGHT);
		rest (1);
	}
}

void crossLinesFade (BITMAP *target) {
	for (z = 0; z < (SCREEN_WIDTH/4+4) / 4; z++) {
		blit (target, screen, z * 8, 0, z * 8, 0, 4, SCREEN_HEIGHT);
		blit (target, screen, (SCREEN_WIDTH-1) - (z * 8), 0, (SCREEN_WIDTH-1) - (z * 8), 0, 4, SCREEN_HEIGHT);
		blit (target, screen, 0, (z * 6), 0, (z * 6), SCREEN_WIDTH, 4);
		blit (target, screen, 0, (SCREEN_HEIGHT - (z * 6)), 0, (SCREEN_HEIGHT - (z * 6)), SCREEN_WIDTH, 4);
		rest (1);
	}
	for (z = (SCREEN_WIDTH/4+4) / 4; z != 0; z--) {
		blit (target, screen, z * 8 + 4, 0, z * 8 + 4, 0, 4, SCREEN_HEIGHT);
		blit (target, screen, (SCREEN_WIDTH-1) - (z * 8) + 4, 0, (SCREEN_WIDTH-1) - (z * 8) + 4, 0, 4, SCREEN_HEIGHT);
		blit (target, screen, 0, (z * 6) + 4, 0, (z * 6) + 4, SCREEN_WIDTH, 4);
		blit (target, screen, 0, (SCREEN_HEIGHT - (z * 6)) + 4, 0, (SCREEN_HEIGHT - (z * 6)) + 4, SCREEN_WIDTH, 4);
		rest (1);
	}
}

void change(BITMAP * target)
{
	show_mouse (NULL);
	if (errorMessage) {
		textout (target, font, errorMessage, errorX, errorY, makecol16 (255, 0, 0));
		errorMessage = NULL;
	}
	switch (rand () % 7) {
		case 0: crossLinesFade (target); break;
		case 1: leftRightWipeFade (target); break;
		case 2: rightLeftWipeFade (target); break;
		case 3: bottomUpWipeFade (target); break;
		case 4: topDownWipeFade (target); break;
		case 5: topDownMultiWipeFade (target); break;
		case 6: spiralPixelFade (target); break;
	}
	clear_keybuf ();
}

void quickChange (BITMAP *target) {
	if (errorMessage) {
		textout (target, font, errorMessage, errorX, errorY, makecol16 (255, 0, 0));
		errorMessage = NULL;
	}
	blit (target, screen, 0, 0, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT);
	rest (1);
}

void flush_inputs()	/* Flush key buffer and waits for but's releases */
{
	do { } while(mouse_b);
	clear_keybuf();
}

void wait_for_input()
{
	do{
#ifdef LINUX
		usleep(1);
#endif /* LINUX */
	} while((!keypressed()) && (!mouse_b));

	flush_inputs();

}

int pickColor (int left, int top, int width, int height, int x, int y)
{
	int r, g, b;
	double value, saturation;
	double hue = ((double)(x - left) / width) * 360;

	double hPos = (double)(y - top) / height;
	if (hPos > 0.5) {
		value = 1.0 - ((hPos - 0.5) * 2);
		saturation = 1.0;
	} else {
		value = 1.0;
		saturation = hPos * 2;
	}

	hsv_to_rgb (hue, saturation, value, &r, &g, &b);

	return (makecol16 (r, g, b));
}

void colorBar (GLOBALDATA *global, ENVIRONMENT *env, int left, int top, int width, int height)
{
	int right = left + width;
	int bottom = top + height;

	for (int x = left; x < right; x++) {
		for (int y = top; y < bottom; y++) {
			putpixel (env->db, x, y, pickColor (left, top, width, height, x, y));
		}
	}
}

int textEntryBox (GLOBALDATA *global, ENVIRONMENT *env, char modify, int x, int y, char *text, unsigned int textLength)
{
	int ke = 0;
	int fontWidth = text_length (font, "Z");
	int fontHeight = text_height (font);
	int leftX = x - (fontWidth * textLength / 2);
	int rightX = x + (fontWidth * textLength / 2);
	int boxWidth = fontWidth * textLength;
	char tempText[textLength + 1];
	int flashCount = 0;
	int lx = mouse_x, ly = mouse_y;

	rectfill (env->db, leftX, y - 2, rightX, y + fontHeight + 2, WHITE);
	rect (env->db, leftX, y - 2, rightX, y + fontHeight + 2, BLACK);
	if (!modify) {
		textout_centre (env->db, font, text, x, y, BLACK);
	}
	env->make_update (leftX - 2, y - 4, fontWidth * textLength + 4, fontHeight + 6);
	env->do_updates ();

	if (!modify) {
		return (boxWidth);
	}
	strncpy (tempText, text, textLength + 1);

	while (((ke >> 8) != KEY_ENTER && (ke >> 8) != KEY_ESC) || strlen (tempText) < 1) {
		int tWidth = text_length (font, tempText);

#ifdef LINUX
		usleep (1);
#endif /* LINUX */

		rectfill (env->db, leftX, y - 2, rightX, y + fontHeight + 2, WHITE);
		rect (env->db, leftX, y - 2, rightX, y + fontHeight + 2, BLACK);
		//rectfill (screen, x - (tWidth / 2), y, x + (tWidth / 2) + 10, y + text_height (font), (flashCount < 5)?WHITE:BLACK);
		textout_centre (env->db, font, tempText, x, y, BLACK);
		rectfill (env->db, x + (tWidth / 2) + 2, y, x + (tWidth / 2) + 10, y + text_height (font), (flashCount < 25)?WHITE:BLACK);
		env->make_update (leftX - 2, y - 4, fontWidth * textLength + 4, fontHeight + 6);
		env->make_update (mouse_x, mouse_y, ((BITMAP *) (global->gfxData.M[0].dat))->w, ((BITMAP *) (global->gfxData.M[0].dat))->h);
		env->make_update (lx, ly, ((BITMAP *) (global->gfxData.M[0].dat))->w, ((BITMAP *) (global->gfxData.M[0].dat))->h);
		lx = mouse_x;
		ly = mouse_y;
		show_mouse (NULL);

		if (keypressed ()) {
			ke = readkey ();
		} else {
			ke = 0;
		}

		if ((ke >> 8) == KEY_BACKSPACE) {
			tempText[strlen (tempText) - 1] = 0;
			rectfill (screen, x - (tWidth / 2), y, x + (tWidth / 2) + 10, y + text_height (font), WHITE);
			env->make_update (x - (tWidth / 2) - 2, y - 2, tWidth + 14, text_height (font) + 4);
		} else if ((ke & 0xff) >= 32 && strlen (tempText) < textLength) {
			tempText[strlen (tempText)] = ke & 0xff;
			tempText[strlen (tempText) + 1] = 0;
			//textprintf (screen, font, x + text_length (font, tempText), y, WHITE, "%c", ke & 0xff);
		} else 
		env->do_updates ();
		show_mouse (screen);
		rest (1);
		flashCount++;
		flashCount = flashCount % 50;
	}
	if ((ke >> 8) != KEY_ESC)
		strncpy (text, tempText, textLength);

//	env->make_update (leftX - 2, y - 4, fontWidth * textLength + 4, fontHeight + 6);
//	env->do_updates ();
	flush_inputs ();

	return (boxWidth);
}

void renderTextLines (ENVIRONMENT *env, char **lines, int scrollOffset)
{
	int yOffset = scrollOffset;
	int lineCount;

	for (lineCount = 0; lines[lineCount] != NULL; lineCount++);

	for (int count = 0; lines[count] != NULL; count++) {
		char *text = &lines[count][2];
		char *header = &lines[count][0];
		int tLen = text_length (font, text);

		switch (header[0]) {
			case '\0':
				yOffset += text_height (font);
				break;
			case 'T':
				textout_centre (env->db, font, text, HALF_WIDTH + 2, HALF_HEIGHT + yOffset + 2, BLACK);
				textout_centre (env->db, font, text, HALF_WIDTH, HALF_HEIGHT + yOffset, WHITE);
				switch (header[1]) {
					case 2:
						yOffset += text_height (font);
						break;
					case 1:
						yOffset += text_height (font) + 1;
						hline (env->db, HALF_WIDTH - tLen / 2, HALF_HEIGHT + yOffset, HALF_WIDTH + tLen / 2, WHITE);
						hline (env->db, HALF_WIDTH - tLen / 2 + 2, HALF_HEIGHT + yOffset + 2, HALF_WIDTH + tLen / 2 + 2, BLACK);
						yOffset += text_height (font) / 2;
						break;
					default:
						yOffset += text_height (font);
						break;
				}
				break;
			case 'L':
				textout_centre (env->db, font, text, HALF_WIDTH, HALF_HEIGHT + yOffset, BLACK);
				yOffset += text_height (font);
				break;
			default:
				break;
		}
		yOffset += text_height (font);
		//yOffset = yOffset % (lineCount * 20);
		//if (yOffset > HALF_HEIGHT)
		//	yOffset = -HALF_HEIGHT + ((yOffset - 1) % 20);
	}
}

void scrollTextList (GLOBALDATA *global, ENVIRONMENT *env, char **lines)
{
	int tOffset = rand ();
	int itemType = rand () % 2;
	int numItems;
	switch (itemType) {
		case 0: // Circles
			numItems = 100;
			break;
		case 1: // Horz lines
			numItems = 30;
			break;
		default:
			numItems = 0;
			break;
	}
	int lineCount;
	for (lineCount = 0; lines[lineCount] != NULL; lineCount++);

	draw_circlesBG (global, env->db, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT);
	drawMenuBackground (global, env, itemType, abs (tOffset), numItems);
	quickChange (env->db);
	set_clip (env->db, HALF_WIDTH - 300, 100,
		HALF_WIDTH + 300, SCREEN_HEIGHT - 100);
	int scrollOffset = 0;
	flush_inputs ();
	show_mouse (NULL);
	do {
#ifdef LINUX
		usleep (1);
#endif /* LINUX */
		drawMenuBackground (global, env, itemType, abs (tOffset), numItems);
		renderTextLines (env, lines, scrollOffset);
		blit (env->db, screen, HALF_WIDTH - 300, 100, HALF_WIDTH - 300, 100, 601, SCREEN_HEIGHT - 199);
		rest (1);
		tOffset++;
		scrollOffset--;
		if (scrollOffset < -(HALF_HEIGHT - 100 + lineCount * 20))
			scrollOffset = HALF_HEIGHT - 100;
	} while (!keypressed () && !mouse_b);
	show_mouse (screen);
	set_clip (env->db, 0, 0, (SCREEN_WIDTH-1), (SCREEN_HEIGHT-1));
	flush_inputs ();
}

void instructions (GLOBALDATA *global, ENVIRONMENT *env)
{
	static char *instructionsList[] = {
		"TInstructions",
		"TOverview",
		"LDestroy the other tanks to gain money,",
		"Lthen use the money to buy bigger, better defenses and",
		"Lweapons to destroy the other tanks with.",
		"",
		"TPlayer Selection",
		"LSelect the players you want in the game,",
		"Lwhite means selected, black unselected.",
		"LThe icons next to the names indicate either a human player (purple circle)",
		"Lor an AI player (green box). The skill of the AI is indicated by the",
		"Lsize of the box (larger is more skilled).",
		"LOnce you've made your selection, click Okay to continue.",
		"",
		"TPurchasing",
		"LUse the mouse to buy items and the following keys to navigate the list:",
		"LScroll Up/Down the list - Up/Down arrow keys",
		"LScroll faster - Page Up/Page Down",
		"",
		"TIn-Game Controls",
		"LUse the following keys to select a weapon or item, aim and fire",
		"LIncrease/Decrease Power - Up/Down arrow keys",
		"LRotate turret Left/Right - Left/Right arrow keys",
		"LSingle-step power and rotation - Hold Ctrl",
		"LNext weapon - Tab",
		"LFire/Activate - Space",
		"",
		"TCreating Players",
		"LSelect the 'Players' button from the main menu.",
		"LSelect the 'Create New' option from the Players menu.",
		"LChoose whether this is an AI or a human using the",
		"L'Type' box (click on the Up/Down arrows).",
		"LChange the name by clicking on it and editing it.",
		"LChange the color by clicking on the color-selection bar.",
		"LClick on Okay.",
		"L10 computer players are automatically created the first",
		"Ltime you run this game.",
		"",
		"TFurther Details",
		"TComputer players",
		"LEach AI player has a personality. They can be defensive, offensive,",
		"Lgo for the cheap stuff all the time or spend it all on the most",
		"Lexpensive thing they can afford.",
		NULL
	};

	scrollTextList (global, env, instructionsList);
}

void credits (GLOBALDATA *global, ENVIRONMENT *env)
{
	char *creditsList[] = {
		"TAtomic Tanks by Stevante Software & Kota 543",
		"TCode & Graphics",
		"LTom Hudson (tom@singular.org.uk)",
		"",
		"TMaintainer",
		"LJuraj Michalek (georgik@host.sk)",
		"",
		"TOriginal Code & Graphics",
		"LSteve Condie (stevante@yahoo.com)",
		"LJuraj Michalek (georgik@host.sk)",
		"",
		"TOriginal Graphics",
		"LTrevor Paulsen (supergass@aol.com)",
		"LRicky Galati (benogalati@yahoo.com)",
		"",
		"TAllegro",
		"LShawn Hargreaves (shawn@talula.demon.co.uk) - http://allegro.sf.net",
		"",
		"LGames for Linux: http://games.linux.sk",
		"LText Logos: http://www.cooltext.com",
		NULL
	};

	scrollTextList (global, env, creditsList);
}

int savePlayers (GLOBALDATA *global, ENVIRONMENT *env, FILE *file)
{
	fwrite (&global->numPermanentPlayers, sizeof (int), 1, file);

	if (global->numPermanentPlayers > 0) {
		for (int count = 0; count < global->numPermanentPlayers; count++) {
			global->allPlayers[count]->saveToFile (file);
		}
	}

	return (ferror (file));
}

int loadPlayers (GLOBALDATA *global, ENVIRONMENT *env, FILE *file)
{
	int numSavedPlayers;

	fread (&numSavedPlayers, sizeof (int), 1, file);

	global->numPermanentPlayers = numSavedPlayers;
	global->allPlayers = (PLAYER**)malloc (sizeof (PLAYER*) * numSavedPlayers);
	if (!global->allPlayers) {
		perror ("atanks.cc: Failed allocating memory for allPlayers in loadPlayers");
		exit (1);
	}
	for (int count = 0; count < numSavedPlayers; count++) {
		global->allPlayers[count] = new PLAYER (global, env, file);
		if (!global->allPlayers[count]) {
			perror ("atanks.cc: Failed allocating memory for players in loadPlayers");
			exit (1);
		}
	}

	return (ferror (file));
}

void newgame (GLOBALDATA *global, ENVIRONMENT *env)
{
	int objCount;
	TANK *tank;

	env->initialise ();
	global->initialise ();
	initialisePlayers (global, env);

	for (objCount = 0; (tank = (TANK*)env->getNextOfClass (TANK_CLASS, &objCount)) && tank; objCount++)
		tank->initialise ();
}

int draw_circlesBG (GLOBALDATA *global, BITMAP *dest, int x, int y, int width, int height)
{
	int largestCircle, circleCount;
	BITMAP *drawTo;

	if (global->cacheCirclesBG)
	{
		if (!global->gfxData.circlesBG)
		{
			global->gfxData.circlesBG = create_bitmap (width, height);
			drawTo = global->gfxData.circlesBG;
		} else {
			blit (global->gfxData.circlesBG, dest, 0, 0, 0, 0, width, height);
			return (0);
		}
	} else {
		drawTo = dest;
	}

	largestCircle = (int)(HALF_WIDTH * (4.0/3.0));
	global->gfxData.circle_gradient_strip = create_gradient_strip (circles_gradient, largestCircle);
	for (circleCount = largestCircle; circleCount > 0; circleCount -= 2)
	circlefill (drawTo, width/2, height/2, circleCount, getpixel (global->gfxData.circle_gradient_strip, 0, largestCircle - circleCount));

	if (global->cacheCirclesBG)
		draw_circlesBG (global, dest, x, y, width, height);

	return (0);
}

ENVIRONMENT *init_game_settings (GLOBALDATA *global)
{
	int count, x, y;
	ENVIRONMENT *env;
	double expSize, disperseSize;

	allegro_init ();
	set_color_depth (16);
	if (set_gfx_mode (screen_mode, SCREEN_WIDTH, SCREEN_HEIGHT, 0, 0) < 0) {
		perror ("set_gfx_mode");
		exit (1);
	}

	if (install_keyboard () < 0) {
		perror ("install_keyboard");
		exit (1);
	}
	if (install_timer () < 0) {
		perror ("install_timer");
		exit (1);
	}
	if (install_mouse () < 0) {
		perror ("install_mouse");
		exit (1);
	}
	if (install_sound (DIGI_AUTODETECT, MIDI_NONE, NULL) < 0) {
		perror ("install_sound");
		exit (1);
	}
	LOCK_VARIABLE(cclock);
	LOCK_FUNCTION(clockadd);
	if (install_int_ex (clockadd, BPS_TO_TIMER (FRAMES_PER_SECOND)) < 0) {
		perror ("install_int_ex");
		exit (1);
	}
	if (install_int (fpsadd, 1000) < 0) {
		perror ("install_int");
		exit (1);
	}

	text_mode (-1);

	srand (time (0));

	WHITE = makecol16 (255, 255, 255);
	BLACK = makecol16 (0, 0, 0);
	PINK = makecol16 (255, 0, 255);

	global->gfxData.TITLE = load_datafile (DATA_DIR "/title.dat");
	if (!global->gfxData.TITLE) {
		perror ("title.dat");
		exit (1);
	}

	show_mouse(NULL);
	blit ((BITMAP *) global->gfxData.TITLE[0].dat, screen, 0, 0, HALF_WIDTH - 320, HALF_HEIGHT - 240, 640, 480);
	show_mouse (screen);

	
	global->gfxData.M = load_datafile (DATA_DIR "/misc.dat");
	if (!global->gfxData.M) {
		perror ("misc.dat");
		exit (1);
	}
	global->gfxData.T = load_datafile (DATA_DIR "/tank.dat");
	if (!global->gfxData.T) {
		perror ("tank.dat");
		exit (1);
	}
	global->gfxData.B = load_datafile (DATA_DIR "/button.dat");
	if (!global->gfxData.B) {
		perror ("button.dat");
		exit (1);
	}
	global->gfxData.TG = load_datafile (DATA_DIR "/tankgun.dat");
	if (!global->gfxData.TG) {
		perror ("tankgun.dat");
		exit (1);
	}
	global->gfxData.MI = load_datafile (DATA_DIR "/missile.dat");
	if (!global->gfxData.MI) {
		perror ("missile.dat");
		exit (1);
	}
	global->gfxData.STOCK_IMAGE = load_datafile(DATA_DIR "/stock.dat");
	if (!global->gfxData.STOCK_IMAGE) {
		perror ("stock.dat");
		exit (1);
	}

	global->SOUND = load_datafile (DATA_DIR "/sound.dat");
	if (!global->SOUND) {
		perror ("sound.dat");
		exit (1);
	}

        for (count = 0; count < LANDS; count++)
          global->gfxData.land_gradient_strips[count] = NULL;
        for (count = 0; count < SKIES; count++)
          global->gfxData.sky_gradient_strips[count] = NULL;

	global->gfxData.explosion_gradient_strip = create_gradient_strip (explosion_gradient, 200);

	expSize = 0;
	disperseSize = 0;
	for (count = 0; count < EXPLOSIONFRAMES; count++)
	{
		global->gfxData.explosions[count] = create_bitmap (214, 214);
		if (count == 0) {
			expSize = 25;
			disperseSize = 0;
		} else if (count < EXPLODEFRAMES - 4)
			expSize += (107 - expSize) / 3;
		else if (count < EXPLODEFRAMES)
			expSize--;
		else if (count == EXPLODEFRAMES)
			disperseSize = 25;
		else
			disperseSize += (107 - disperseSize) / 2;

		clear_to_color (global->gfxData.explosions[count], PINK);
		for (y = (int)expSize; y > disperseSize; y--) {
			double value;
			value = pow ((double)y / expSize, count / 4 + 1);
			circlefill (global->gfxData.explosions[count], 107, 107, y, getpixel (global->gfxData.explosion_gradient_strip, 0, (int)(value * 200)));
		}
		if (disperseSize)
			circlefill (global->gfxData.explosions[count], 107, 107, (int)disperseSize, PINK);
	}

	expSize = 0;
	disperseSize = 0;
	for (count = 0; count < EXPLOSIONFRAMES; count++)
	{
		global->gfxData.flameFront[count] = create_bitmap (600, 30);
		if (count == 0) {
			expSize = 10;
			disperseSize = 0;
		} else if (count < EXPLODEFRAMES - 4)
			expSize += (300 - expSize) / 3;
		else if (count < EXPLODEFRAMES)
			expSize--;
		else if (count == EXPLODEFRAMES)
			disperseSize = 10;
		else
			disperseSize += (300 - disperseSize) / 2;

		clear_to_color (global->gfxData.flameFront[count], PINK);
		for (y = (int)expSize; y > disperseSize; y--) {
			double value;
			value = pow ((double)y / expSize, count / 4 + 1);
			ellipsefill (global->gfxData.flameFront[count], 300, 15, y, y / 20, getpixel (global->gfxData.explosion_gradient_strip, 0, (int)(value * 200)));
		}
		if (disperseSize)
			ellipsefill (global->gfxData.flameFront[count], 300, 15, (int)disperseSize, (int)disperseSize / 16, PINK);
	}

	global->gfxData.topbar = create_bitmap (SCREEN_WIDTH, MENU);
	global->gfxData.topbar_gradient_strip = create_gradient_strip (topbar_gradient, 100);
	if (!global->ditherGradients) {
		for (count = 0; count < MENU; count++) {
			float adjCount = (100.0 / MENU) * count;
			line (global->gfxData.topbar, 0, count, SCREEN_WIDTH - 1, count, getpixel (global->gfxData.topbar_gradient_strip, 0, (int)adjCount));
		}
	} else {
		for (x = 0; x < SCREEN_WIDTH; x++) {
		        for (y = 0; y < MENU; y++) {
				float adjY = (100.0 / MENU) * y;
				int offset;
				if ((adjY < 2) || (adjY > 100 - 2))
					offset = 0;
				else
					offset = rand () % 4 - 2;
				putpixel (global->gfxData.topbar, x, y, getpixel (global->gfxData.topbar_gradient_strip, 0, (int)adjY + offset));
			}
		}
	}

	global->gfxData.stuff_bar[0] = create_bitmap (STUFF_BAR_WIDTH, STUFF_BAR_HEIGHT);
	global->gfxData.stuff_bar[1] = create_bitmap (STUFF_BAR_WIDTH, STUFF_BAR_HEIGHT);
	global->gfxData.stuff_icon_base = create_bitmap (STUFF_BAR_WIDTH/10, STUFF_BAR_HEIGHT);
	clear_to_color (global->gfxData.stuff_bar[0], PINK);
	clear_to_color (global->gfxData.stuff_bar[1], PINK);
	clear_to_color (global->gfxData.stuff_icon_base, PINK);
	global->gfxData.stuff_bar_gradient_strip = create_gradient_strip (stuff_bar_gradient, STUFF_BAR_WIDTH);
	for (x = 0; x < STUFF_BAR_WIDTH; x++)
	{
		for (y = 0; y < STUFF_BAR_HEIGHT; y++)
		{
			double sides_dist = 0.1, circle_dist;
			circle_dist = vector_length_f ((float)x-(STUFF_BAR_WIDTH - 75), (float)y - (STUFF_BAR_HEIGHT/2 - 2), 0);
			if (circle_dist < 75)
				circle_dist = 1 - (circle_dist / 75.0);
			else
				circle_dist = 0;

			if (x < (STUFF_BAR_HEIGHT/2 - 2))
				sides_dist -= 0.1 - ((float)x / 150.0);
			else if (x > STUFF_BAR_WIDTH - (STUFF_BAR_HEIGHT/2 - 2))
				sides_dist -= ((float)(x - (STUFF_BAR_WIDTH - (STUFF_BAR_HEIGHT/2 - 2))) / 150.0);

			if (y < STUFF_BAR_HEIGHT/2 - 2)
				sides_dist -= 0.1 - ((float)(y) / 150.0);
			else
				sides_dist -= ((float)(y - (STUFF_BAR_HEIGHT/2 - 2)) / 150.0);

			sides_dist -= circle_dist * circle_dist;
			if (sides_dist > ((double)x / 1000.0))
				sides_dist = ((double)x / 1000.0);
			if (sides_dist < 0)
				sides_dist = 0;
			if (circle_dist > 1)
				circle_dist = 1;

			if (x < STUFF_BAR_WIDTH/10)
				putpixel (global->gfxData.stuff_icon_base, x, y, getpixel (global->gfxData.stuff_bar_gradient_strip, 0, (int)((sides_dist + circle_dist) * (STUFF_BAR_WIDTH-1))));

			if (y < STUFF_BAR_HEIGHT - 5) {
				putpixel (global->gfxData.stuff_bar[0], x, y, getpixel (global->gfxData.stuff_bar_gradient_strip, 0, (int)((sides_dist + circle_dist) * (STUFF_BAR_WIDTH-1))));
				putpixel (global->gfxData.stuff_bar[1], x, y, getpixel (global->gfxData.stuff_bar_gradient_strip, 0, (int)((sides_dist + circle_dist + 0.05) * (STUFF_BAR_WIDTH-1))));
			}
		}
	}

	set_mouse_sprite ((BITMAP *) global->gfxData.M[0].dat);
	set_mouse_sprite_focus (0, 0);

	srand (time (0));
	global->window.x = 0;
	global->window.y = 0;
	global->window.w = 0;
	global->window.h = 0;
	for (z = 0; z < MAXUPDATES; z++) {
		global->updates[z].x = 0;
		global->updates[z].y = 0;
		global->updates[z].w = 0;
		global->updates[z].h = 0;
	}

	env = new ENVIRONMENT (global);
	if (!env) {
		perror ("atanks.cc: Allocating env in init_game_settings");
		exit (1);
	}

	clear_to_color (env->db, BLACK);

	return (env);
}

class BUTTON
{
	private:
		GLOBALDATA *_global;
		ENVIRONMENT *_env;

	public:
		BOX location;
		int xl, yl;
		char *text;
		BITMAP *bmp;
		BITMAP *hover;
		BITMAP *depressed;
		SAMPLE *click;

		BUTTON (GLOBALDATA *global, ENVIRONMENT *env, int x1, int y1, char *text1, BITMAP *bmp1, BITMAP *hover1, BITMAP *depressed1);
	int	isPressed (); //if button pressed returns 1
	int	isMouseOver (); //Cursor is over button
	void	draw (BITMAP *dest);
};

BUTTON::BUTTON(GLOBALDATA *global, ENVIRONMENT *env, int x1, int y1, char *text1, BITMAP *bmp1, BITMAP *hover1, BITMAP *depressed1)
{
	_global = global;
	_env = env;
	location.x = x1;
	location.y = y1;
	text = text1;
	click = (SAMPLE *)_global->SOUND[8].dat;
	bmp = bmp1;
	hover = hover1;
	depressed = depressed1;
	location.w = bmp->w;
	location.h = bmp->h;
	xl = location.x + bmp->w;
	yl = location.y + bmp->h;
}

void BUTTON::draw (BITMAP *dest)
{
	draw_sprite (dest, (BITMAP *)(isMouseOver())?((isPressed())?depressed:hover):bmp, location.x, location.y);
	if (text)
		textout_centre (dest, font, text, location.x+75, location.y + 6, WHITE);
	_env->make_update (location.x, location.y, xl, yl);
}

int BUTTON::isMouseOver ()
{
	if ((mouse_x >= location.x) &&
		(mouse_y >= location.y) &&
		(mouse_x <= xl) &&
		(mouse_y <= yl)) {
		return (1);
	} else {
		return (0);
	}
}
int BUTTON::isPressed()
{
	if ((mouse_b == 1) && isMouseOver ()) {
		play_sample (click, 128, 128, 1000, 0);
		return 1;
	} else {
		return 0;
	}
}

#include "menu.h"
int options (GLOBALDATA *global, ENVIRONMENT *env, MENUDESC *menu);
int createNewPlayer (GLOBALDATA *global, ENVIRONMENT *env, void *data)
{
	PLAYER *newPlayer = global->createNewPlayer (env);
	if (!newPlayer) {
		perror ("atanks.cc: Failed allocating memory in createNewPlayer");
		exit (1);
	}
	options (global, env, (MENUDESC*)newPlayer->menudesc);
	return (-1);
}

int destroyPlayer (GLOBALDATA *global, ENVIRONMENT *env, void *data)
{
	int optionsRetVal;
	char sureMessage[200];
	PLAYER *tempPlayer = (PLAYER*)data;
	MENUDESC areYouSureMenu = {"Are You Sure?", 0, NULL, TRUE, TRUE};

	sprintf (sureMessage, "This player (%s) will be permanently deleted", tempPlayer->getName ());
	errorMessage = sureMessage;
	errorX = HALF_WIDTH - text_length (font, errorMessage) / 2;
	errorY = 170;
	optionsRetVal = options (global, env, &areYouSureMenu);
	if (optionsRetVal >> 8 != KEY_ESC) {
		global->destroyPlayer (tempPlayer);
		return (-2);
	}
	return (KEY_SPACE << 8);
}

int displayPlayerName (GLOBALDATA *global, ENVIRONMENT *env, int x, int y, void *data)
{
	PLAYER *player = (PLAYER*)data;
	char *name = player->getName ();
	int textHeight = text_height (font);
	int textLength = text_length (font, name);

	if ((int)player->type == HUMAN_PLAYER) {
		circlefill (env->db, x - textLength - textHeight / 2 - 2,
				y + textHeight / 2,
				textHeight / 2, makecol16 (200, 100, 255));
		circle (env->db, x - textLength - textHeight / 2 - 2,
				y + textHeight / 2,
				textHeight / 2, BLACK);
	} else {
		rectfill (env->db, x - textLength - 10,
			y + textHeight - 2 - (textHeight / LAST_PLAYER_TYPE * (int)player->type),
			x - textLength - 2, y + textHeight - 2,
			makecol16 (100, 255, 100));
		rect (env->db, x - textLength - 10,
			y + textHeight - 2 - (textHeight / LAST_PLAYER_TYPE * (int)player->type),
			x - textLength - 2, y + textHeight - 2,
			BLACK);
		for (int lineCount = 1; lineCount < player->type; lineCount++) {
			hline (env->db, x - textLength - 10,
				y + textHeight - 2 - (textHeight / LAST_PLAYER_TYPE * lineCount),
				x - textLength - 2, BLACK);
		}
	}
	textout (env->db, font, name, x - textLength, y, player->color);

	return (0);
}

int options (GLOBALDATA *global, ENVIRONMENT *env, MENUDESC *menu)
{
	MENUENTRY *opts;
	int numEntries;
	char *title;
	int ke;
	int mouseLeftPressed;
#include "menucontent.h"

	if (!menu)
		menu = &mainMenu;

	opts = menu->entries;
	numEntries = menu->numEntries;
	title = menu->title;

	char buff[20];
	int done, updateoption[numEntries], lb;

	int stop = 0;

	BUTTON *but_okay = NULL, *but_quit = NULL;
	if (menu->okayButton) {
		int xpos = HALF_WIDTH - 80;
		if (menu->quitButton)
			xpos -= 80;
		but_okay = new BUTTON (global, env, xpos, HALF_HEIGHT + 160, "Okay", (BITMAP*)global->gfxData.M[7].dat, (BITMAP*)global->gfxData.M[7].dat, (BITMAP*)global->gfxData.M[8].dat);
		if (!but_okay) {
			perror ("atanks.cc: Failed allocating memory for but_okay in options");
			exit (1);
		}
	}
	if (menu->quitButton) {
		int xpos = HALF_WIDTH - 80;
		if (menu->okayButton)
			xpos += 80;
		but_quit = new BUTTON (global, env, xpos, HALF_HEIGHT + 160, "Back", (BITMAP*)global->gfxData.M[7].dat, (BITMAP*)global->gfxData.M[7].dat, (BITMAP*)global->gfxData.M[8].dat);
		if (!but_quit) {
			perror ("atanks.cc: Failed allocating memory for but_quit in options");
			exit (1);
		}
	}

	mouseLeftPressed = done = lb = env->mouseclock = cclock = 0;
	fi = 1;
	
	for (z = 0; z < numEntries; z++) {
		updateoption[z] = 1;
	}
	
	flush_inputs ();

	do {
#ifdef LINUX
		usleep (1);
#endif /* LINUX */
		while (cclock > 0) {
			cclock--;
			if (!lb && mouse_b & 1) {
				env->mouseclock = 0;
				mouseLeftPressed = 1;
			} else {
				mouseLeftPressed = 0;
			}
			lb = (mouse_b & 1) ? 1 : 0;
			if (mouse_b & 1 && !env->mouseclock) {
				for (z = 0; z < numEntries; z++) {
					int midX = opts[z].x;
					int midY = opts[z].y;
					if (opts[z].type == OPTION_MENUTYPE) {
						sprintf (buff, "-> %s", opts[z].name);
						if ((!opts[z].viewonly) && mouse_x > midX - text_length (font, buff) && mouse_x < midX && mouse_y >= midY && mouse_y < midY + 10) {
							int optsRetVal = options (global, env, (MENUDESC*)opts[z].value);
							if (optsRetVal < 0) {
								return (optsRetVal + 1);
							}
							fi = 1;
							for (z = 0; z < numEntries; z++) {
								updateoption[z] = 1;
							}
						}
					} else if (opts[z].type == OPTION_ACTIONTYPE) {
						sprintf (buff, "-> %s", opts[z].name);
						if ((!opts[z].viewonly) && mouse_x > midX - text_length (font, buff) && mouse_x < midX && mouse_y >= midY && mouse_y < midY + 10) {
							int (*action) (GLOBALDATA*, ENVIRONMENT*, void*) = (int (*)(GLOBALDATA*, ENVIRONMENT*, void*))opts[z].value;
							int actionRetVal = action (global, env, opts[z].data);
							if (actionRetVal)
								return (actionRetVal);
						}
					} else if (opts[z].type == OPTION_TEXTTYPE) {
						int boxWidth = textEntryBox (global, env, FALSE, midX + 50, midY, (char*)opts[z].value, 10);
						if ((!opts[z].viewonly) && mouse_x > midX - text_length (font, buff) && mouse_x < midX + 50 + boxWidth && mouse_y >= midY && mouse_y < midY + 10) {
							textEntryBox (global, env, TRUE, midX + 50, midY, (char*)opts[z].value, 10);
							updateoption[z] = 1;
						}
					} else if (opts[z].type == OPTION_COLORTYPE) {
						if ((!opts[z].viewonly) && mouse_x > midX && mouse_x < midX + 100 && mouse_y >= midY && mouse_y < midY + 15) {
							*(int*)opts[z].value = pickColor (midX, midY, 100, 15, mouse_x, mouse_y);
							updateoption[z] = 1;
						}
						colorBar (global, env, midX, midY, 100, 15);
						rectfill (env->db, midX + 110, midY, midX + 130, midY + 10, *(int*)opts[z].value);
						rect (env->db, midX + 110, midY, midX + 130, midY + 10, BLACK);
					} else if (opts[z].type == OPTION_TOGGLETYPE) {
						int tlen = text_length (font, buff);
						int thgt = text_height (font);
						if (mouseLeftPressed && (!opts[z].viewonly) && mouse_x > midX - tlen / 2 && mouse_x < midX + tlen / 2 && mouse_y >= midY && mouse_y < midY + thgt) {
							if (*opts[z].value == 0)
								*opts[z].value = 1;
							else
								*opts[z].value = 0;
							mouseLeftPressed = 1;
							updateoption[z] = 1;
						}
					} else {
						if (!opts[z].viewonly) {
							if (mouse_x >= midX + 100 && mouse_x < midX + 110 && mouse_y >= midY && mouse_y < midY + 10) {
								*opts[z].value -= opts[z].increment;
								updateoption[z] = 1;
							}
							if (mouse_x >= midX + 112 && mouse_x < midX + 122 && mouse_y >= midY && mouse_y < midY + 10) {
								*opts[z].value += opts[z].increment;
								updateoption[z] = 1;
							}
							/*if (mouse_x >= midX + 134 && mouse_x < midY + 154 && mouse_y >= midY && mouse_y < midY + 10) {
								*opts[z].value = opts[z].defaultv;
								updateoption[z] = 1;
							}*/
							if (*opts[z].value > opts[z].max) {
								*opts[z].value = opts[z].max;
							}
							if (*opts[z].value < opts[z].min) {
								*opts[z].value = opts[z].min;
							}
						}
					}
				}
			}
			env->mouseclock++;
			if (env->mouseclock > 10) {
				env->mouseclock = 0;
			}
		}
		
		env->make_update (mouse_x, mouse_y, ((BITMAP *) (global->gfxData.M[0].dat))->w, ((BITMAP *) (global->gfxData.M[0].dat))->h);
		env->make_update (lx, ly, ((BITMAP *) (global->gfxData.M[0].dat))->w, ((BITMAP *) (global->gfxData.M[0].dat))->h);
		lx = mouse_x;
		ly = mouse_y;
		show_mouse (NULL);
		
		if (fi) {
			drawMenuBackground (global, env, 0, rand (), 400);
			textout (env->db, font, title, HALF_WIDTH - 3 - text_length (font, title), 150, BLACK);
			textout (env->db, font, title, HALF_WIDTH - 5 - text_length (font, title), 148, WHITE);
			for (z = 0; z < numEntries; z++) {
				int midX = opts[z].x;
				int midY = opts[z].y;
				if (opts[z].type == OPTION_TOGGLETYPE) {
					int color = (*opts[z].value)?WHITE:BLACK;
					ellipsefill (env->db, midX, midY + text_height (font) / 2, text_length (font, opts[z].name) / 2, text_height (font), color);
				}
				if (opts[z].displayFunc) {
					if (opts[z].type == OPTION_TOGGLETYPE) {
						opts[z].displayFunc (global, env,
								midX + text_length (font, opts[z].name) / 2, midY,
								opts[z].data);
					} else {
						opts[z].displayFunc (global, env,
								midX, midY,
								opts[z].data);
					}
				} else if (opts[z].type == OPTION_MENUTYPE) {
					sprintf (buff, "-> %s", opts[z].name);
					textout (env->db, font, buff, midX - text_length (font, buff), midY, opts[z].color);
				} else if (opts[z].type == OPTION_ACTIONTYPE) {
					sprintf (buff, "-> %s", opts[z].name);
					textout (env->db, font, buff, midX - text_length (font, buff), midY, opts[z].color);
				} else if (opts[z].type == OPTION_TEXTTYPE) {
					sprintf (buff, "%s:", opts[z].name);
					textout (env->db, font, buff, midX - text_length (font, buff), midY, opts[z].color);
				} else if (opts[z].type == OPTION_COLORTYPE) {
					sprintf (buff, "%s:", opts[z].name);
					textout (env->db, font, buff, midX - text_length (font, buff), midY, opts[z].color);
				} else if (opts[z].type == OPTION_TOGGLETYPE) {
					sprintf (buff, "%s", opts[z].name);
					textout_centre (env->db, font, buff, midX, midY, opts[z].color);
				} else {
					sprintf (buff, "%s:", opts[z].name);
					textout (env->db, font, buff, midX - text_length (font, buff), midY, opts[z].color);
					if (!opts[z].viewonly) {
						draw_sprite_v_flip (env->db, (BITMAP *) global->gfxData.M[6].dat, midX + 100, midY);
						draw_sprite (env->db, (BITMAP *) global->gfxData.M[6].dat, midX + 112, midY);
					}
				}
			}
			if (but_okay) but_okay->draw (env->db);
			if (but_quit) but_quit->draw (env->db);
		}
		
		for (z = 0; z < numEntries; z++) {
			int midX = opts[z].x;
			int midY = opts[z].y;
			if (updateoption[z]) {
				updateoption[z] = 0;
				if (opts[z].type == OPTION_TOGGLETYPE) {
					int color = (*opts[z].value)?WHITE:BLACK;
					ellipsefill (env->db, midX, midY + text_height (font) / 2, text_length (font, opts[z].name) / 2, text_height (font), color);
				}
				if (opts[z].displayFunc) {
					if (opts[z].type == OPTION_TOGGLETYPE) {
						opts[z].displayFunc (global, env,
								midX + text_length (font, opts[z].name) / 2, midY,
								opts[z].data);
					} else {
						opts[z].displayFunc (global, env,
								midX, midY,
								opts[z].data);
					}
					env->make_update (midX - 100, midY - text_height (font), 250, 20);
				} else if (opts[z].type != OPTION_MENUTYPE && opts[z].type != OPTION_ACTIONTYPE) {
					if (opts[z].type == OPTION_DOUBLETYPE) {
						char buff[10];
						sprintf (buff, opts[z].format, *opts[z].value);
						textEntryBox (global, env, FALSE, midX + 50, midY, buff, 10);
					} else if (opts[z].type == OPTION_TEXTTYPE) {
						textEntryBox (global, env, FALSE, midX + 50, midY, (char*)opts[z].value, 10);
					} else if (opts[z].type == OPTION_COLORTYPE) {
						colorBar (global, env, midX, midY, 100, 15);
						rectfill (env->db, midX + 110, midY, midX + 130, midY + 10, *(int*)opts[z].value);
						rect (env->db, midX + 110, midY, midX + 130, midY + 10, BLACK);
					} else if (opts[z].type == OPTION_TOGGLETYPE) {
						sprintf (buff, "%s", opts[z].name);
						textout_centre (env->db, font, buff, midX, midY, opts[z].color);
					} else if (opts[z].specialOpts) {
						textEntryBox (global, env, FALSE, midX + 50, midY, opts[z].specialOpts[(int) *opts[z].value], 10);
					}
					env->make_update (midX - 100, midY - 2, 250, 20);
				}
			}
		}

		if (fi) {
			fi = 0;
			quickChange(env->db);
		}

		if(but_quit && but_quit->isPressed()) {
			global->command = GLOBAL_COMMAND_MENU;
			stop = 1;
		}
		if(but_okay && but_okay->isPressed()) {
			stop = 2;
		}

		if (but_okay) but_okay->draw (env->db);
		if (but_quit) but_quit->draw (env->db);

		show_mouse(env->db);
		env->do_updates ();
	} while ((!keypressed ()) && (!stop) );
	if (!stop)
		ke = readkey ();
	else if (stop == 2)
		ke = KEY_SPACE << 8;
	else
		ke = KEY_ESC << 8;
	
	flush_inputs();

	if (but_quit)
		delete but_quit;
	if (but_okay)
		delete but_okay;

	return (ke);
}

int editPlayers (GLOBALDATA *global, ENVIRONMENT *env)
{
	int optionsRetVal;
	int rows = (SCREEN_HEIGHT - 400) / 15;
	int columns = (global->numPermanentPlayers / rows) + 1;
	rows = (rows / columns) + 1;

	MENUENTRY *playersOpts;
	MENUDESC playersMenu;
	playersOpts = new MENUENTRY[1 + global->numPermanentPlayers];
	if (!playersOpts) {
		perror ("atanks.cc: Failed allocating memory for playersOpts in editPlayers");
		exit (1);
	}
	playersOpts[0].name = "Create New";
	playersOpts[0].displayFunc = NULL;
	playersOpts[0].color = WHITE;
	playersOpts[0].value = (double*)createNewPlayer;
	playersOpts[0].data = NULL;
	playersOpts[0].type = OPTION_ACTIONTYPE;
	playersOpts[0].viewonly = FALSE;
	playersOpts[0].x = HALF_WIDTH - 3;
	playersOpts[0].y = HALF_HEIGHT - 68 - 15;

	playersMenu.title = "Players";
	playersMenu.numEntries = 1 + global->numPermanentPlayers;
	playersMenu.entries = playersOpts;
	playersMenu.quitButton = TRUE;
	playersMenu.okayButton = FALSE;

	for (int count = 0; count < global->numPermanentPlayers; count++) {
		MENUENTRY *opt = &playersOpts[1 + count];

		opt->name = global->allPlayers[count]->getName ();
		opt->displayFunc = displayPlayerName;
		opt->data = global->allPlayers[count];
		opt->color = global->allPlayers[count]->color;
		opt->value = (double*)global->allPlayers[count]->menudesc;
		opt->type = OPTION_MENUTYPE;
		opt->viewonly = FALSE;
		opt->x = HALF_WIDTH - (((count % columns) - (columns / 2)) * 90) - (((columns + 1) % 2) * 45);
		opt->y = HALF_HEIGHT - 68 + ((count / columns) * 15);
	}
	optionsRetVal = options (global, env, &playersMenu);

	delete playersOpts;

	return (optionsRetVal);
}

int selectPlayers (GLOBALDATA *global, ENVIRONMENT *env)
{
	MENUENTRY roundOpt = {"Rounds", NULL, WHITE, (double*)&global->rounds, NULL, "%2.0f", 1, 100, 1, 5, NULL, OPTION_DOUBLETYPE, FALSE, HALF_WIDTH - 3, 200};
	int optionsRetVal;
	int rows = (SCREEN_HEIGHT - 400) / 15;
	int columns = (global->numPermanentPlayers / rows) + 1;
	rows = (rows / columns) + 1;

	MENUENTRY *playersOpts;
	MENUDESC playersMenu;
	playersOpts = new MENUENTRY[global->numPermanentPlayers + 1];
	if (!playersOpts) {
		perror ("atanks.cc: Failed allocating memory for playersOpts in selectPlayers");
		exit (1);
	}

	playersMenu.title = "Select Players";
	playersMenu.numEntries = global->numPermanentPlayers + 1;
	playersMenu.entries = playersOpts;
	playersMenu.quitButton = TRUE;
	playersMenu.okayButton = TRUE;

	for (int count = 0; count < global->numPermanentPlayers; count++) {
		MENUENTRY *opt = &playersOpts[count];

		opt->name = global->allPlayers[count]->getName ();
		opt->displayFunc = displayPlayerName;
		opt->data = global->allPlayers[count];
		opt->color = global->allPlayers[count]->color;
		opt->value = (double*)&global->allPlayers[count]->selected;
		opt->type = OPTION_TOGGLETYPE;
		opt->viewonly = FALSE;
		opt->x = HALF_WIDTH - (((count % columns) - (columns / 2)) * 90) - (((columns + 1) % 2) * 45);
		opt->y = HALF_HEIGHT - 68 + ((count / columns) * 15);
	}
	memcpy (&playersOpts[global->numPermanentPlayers], &roundOpt, sizeof (MENUENTRY));

	do {
		optionsRetVal = options (global, env, &playersMenu);
		if (optionsRetVal >> 8 != KEY_ESC) {
			int playerCount = 0;
			global->numPlayers = 0;
			for (z = 0; z < global->numPermanentPlayers; z++) {
				if (global->allPlayers[z]->selected) {
					global->addPlayer (global->allPlayers[z]);
					playerCount++;
				}
			}
			if ((playerCount < 2) || (playerCount > MAXPLAYERS)) {
				optionsRetVal = -1;
				if (playerCount < 2)
					errorMessage = "You must select at least 2 players";
				else if (playerCount > 10)
					errorMessage = "You can only have up to 10 players";
				errorX = HALF_WIDTH - text_length (font, errorMessage) / 2;
				errorY = 170;
			} else {
				optionsRetVal = 0;
			}
		}
	} while ((optionsRetVal != 0) && (optionsRetVal >> 8 != KEY_ESC));

	delete playersOpts;

	return (optionsRetVal);
}

void title (GLOBALDATA *global)
{
	show_mouse(NULL);
	blit ((BITMAP *) global->gfxData.TITLE[0].dat, screen, 0, 0, HALF_WIDTH - 320, HALF_HEIGHT - 240, 640, 480);
	//show_mouse (screen);
	clear_keybuf ();

	//wait_for_input();
	
	//show_mouse (NULL);
}

int gradientColorPoint (gradient *grad, double length, double line)
{
	int pointCount = 0;
	double point = line / length;
	int color;

	for (pointCount = 0; (point >= grad[pointCount].point) && (grad[pointCount].point != -1); pointCount++);
	pointCount--;

	if (pointCount == -1) {
		color = makecol16 (grad[0].color.r, grad[0].color.g, grad[0].color.b);
	} else if (grad[pointCount + 1].point == -1) {
		color = makecol16 (grad[pointCount].color.r, grad[pointCount].color.g, grad[pointCount].color.b);
	} else {
		double i = (point - grad[pointCount].point) / (grad[pointCount + 1].point - grad[pointCount].point);
		int r = (int)(interpolate (grad[pointCount].color.r, grad[pointCount + 1].color.r, i));
		int g = (int)(interpolate (grad[pointCount].color.g, grad[pointCount + 1].color.g, i));
		int b = (int)(interpolate (grad[pointCount].color.b, grad[pointCount + 1].color.b, i));

		color = makecol16 (r, g, b);
	}

	return (color);
}

int menu (GLOBALDATA *global, ENVIRONMENT *env)
{
	int bf = 0, bfdd = 1, ban, anclock, lb, updateplayers, done, updaterounds;
	BUTTON but_play(global, env, HALF_WIDTH - 75, HALF_HEIGHT - 235, NULL, (BITMAP*)global->gfxData.B[0].dat, (BITMAP*)global->gfxData.B[0].dat, (BITMAP*)global->gfxData.B[1].dat);
	BUTTON but_help(global, env, HALF_WIDTH - 75, HALF_HEIGHT - 185, NULL, (BITMAP*)global->gfxData.B[2].dat, (BITMAP*)global->gfxData.B[2].dat, (BITMAP*)global->gfxData.B[3].dat);
	BUTTON but_options(global, env, HALF_WIDTH - 75, HALF_HEIGHT - 135, NULL, (BITMAP*)global->gfxData.B[4].dat, (BITMAP*)global->gfxData.B[4].dat, (BITMAP*)global->gfxData.B[5].dat);
	BUTTON but_players(global, env, HALF_WIDTH - 75, HALF_HEIGHT - 85, NULL, (BITMAP*)global->gfxData.B[6].dat, (BITMAP*)global->gfxData.B[6].dat, (BITMAP*)global->gfxData.B[7].dat);
	BUTTON but_credits(global, env, HALF_WIDTH - 75, HALF_HEIGHT - 35, NULL, (BITMAP*)global->gfxData.B[8].dat, (BITMAP*)global->gfxData.B[8].dat, (BITMAP*)global->gfxData.B[9].dat);
	BUTTON but_quit(global, env, HALF_WIDTH - 75, HALF_HEIGHT + 15, NULL, (BITMAP*)global->gfxData.B[10].dat, (BITMAP*)global->gfxData.B[10].dat, (BITMAP*)global->gfxData.B[11].dat);

	BUTTON *button[MENUBUTTONS] = {&but_play, &but_help, &but_options,
					&but_players, &but_credits, &but_quit};

	ban = -1;
	cclock = global->updateCount = lx = ly = anclock = env->mouseclock = 0;
	lb = updateplayers = done = updaterounds = 0;
	fi = global->stopwindow = 1;
	while (!done) {

#ifdef LINUX
		usleep (1);

#endif /* LINUX */
		while (cclock > 0) {
			cclock--;
			zz = 0;
			for (z = 0; z < MENUBUTTONS; z++) {
				if (button[z]->isMouseOver ()) {
					if (ban > -1 && ban != z) {
						button[z]->draw (env->db);
						//env->make_update (button[ban]->location.x, button[ban]->location.y, button[ban]->location.w, button[ban]->location.h);
					}
					ban = z;
					zz = 1;
					/*anclock++;
					if (anclock > 6) {
						anclock = 0;
						bf += bfdd;
						if (bf > BUTTONFRAMES - 2)
							bfdd = -1;
						if (bf < 2)
							bfdd = 1;
					}*/
					break;
				}
			}
			if (!zz) {
				bf = 0;
				bfdd = 1;
			}
			if (!lb && mouse_b & 1)
				env->mouseclock = 0;
			lb = (mouse_b & 1) ? 1 : 0;
			if (mouse_b & 1) {
				for (z = 0; z < MENUBUTTONS; z++) {
					if (button[z]->isPressed ()) {
						if (z == 0)
							done = 1;
						if (z == 1)
							global->command = GLOBAL_COMMAND_HELP, done = 1;
						if (z == 2)
							global->command = GLOBAL_COMMAND_OPTIONS, done = 1;
						if (z == 3)
							global->command = GLOBAL_COMMAND_PLAYERS, done = 1;
						if (z == 4)
							global->command = GLOBAL_COMMAND_CREDITS, done = 1;
						if (z == 5)
							global->command = GLOBAL_COMMAND_QUIT, done = 1;
					}
				}
				//if (global->rounds > 1 && !env->mouseclock
				//    && mouse_x >= HALF_WIDTH - 60 && mouse_x < HALF_WIDTH - 50 && mouse_y >= HALF_HEIGHT + 199 && mouse_y < HALF_HEIGHT + 209) {
				//	updaterounds = 1;
				//	global->rounds--;
				//}
				//if (global->rounds < 100 && !env->mouseclock
				//    && mouse_x >= HALF_WIDTH + 64 && mouse_x < HALF_WIDTH + 74 && mouse_y >= HALF_HEIGHT + 199 && mouse_y < HALF_HEIGHT + 209) {
				//	updaterounds = 1;
				//	global->rounds++;
				//}
			}
			env->mouseclock++;
			if (env->mouseclock > 10)
				env->mouseclock = 0;
		}
		if (updaterounds) {
			updaterounds = 0;
			env->make_update (HALF_WIDTH + 27, HALF_HEIGHT + 198, 32, 32);
		}
		show_mouse (NULL);
		if (fi) {
			draw_circlesBG (global, env->db, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT);
			//textout (env->db, font, "Rounds: ", HALF_WIDTH - 45, HALF_HEIGHT + 200, BLACK);
			//draw_sprite_v_flip (env->db, (BITMAP *) global->gfxData.M[6].dat, HALF_WIDTH - 60, HALF_HEIGHT + 199);
			//draw_sprite (env->db, (BITMAP *) global->gfxData.M[6].dat, HALF_WIDTH + 64, HALF_HEIGHT + 199);
			for (z = 0; z < MENUBUTTONS; z++) {
				button[z]->draw (env->db);
			}
		}
		if (ban > -1) {
			button[ban]->draw (env->db);
			//env->make_update (button[ban]->location.x, button[ban]->location.y, button[ban]->location.w, button[ban]->location.h);
		}
		//rectfill (env->db, HALF_WIDTH + 27, HALF_HEIGHT + 198, HALF_WIDTH + 59, HALF_HEIGHT + 210, WHITE);
		//rect (env->db, HALF_WIDTH + 27, HALF_HEIGHT + 198, HALF_WIDTH + 59, HALF_HEIGHT + 210, BLACK);
		//textprintf_centre (env->db, font, HALF_WIDTH + 43, HALF_HEIGHT + 200, BLACK, "%d", global->rounds);
		show_mouse (env->db);
		env->make_update (mouse_x, mouse_y, ((BITMAP *) (global->gfxData.M[0].dat))->w, ((BITMAP *) (global->gfxData.M[0].dat))->h);
		env->make_update (lx, ly, ((BITMAP *) (global->gfxData.M[0].dat))->w, ((BITMAP *) (global->gfxData.M[0].dat))->h);
		lx = mouse_x;
		ly = mouse_y;
		if ((key[KEY_Q]) || (key[KEY_F10])) {
			return SIG_QUIT_GAME;
		}
		if (fi) {
			fi = 0;
			change (env->db);
		}
		env->do_updates ();
	}
	clear_keybuf ();
	return SIG_OK;
}

void draw_text_in_box (ENVIRONMENT *env, BOX *region, char *text)
{
	char buffer[1024];
	unsigned int lineBegin;
	int lastSpace = 0;
	int lineCount;
	int charCount;
	int buffCount ;

	rectfill (env->db, region->x, region->y, region->w, region->h,
			makecol16 (0,0,128));
	rect     (env->db, region->x, region->y, region->w, region->h,
			makecol16 (128,128,255));

	lineBegin = 0;
	lineCount = 0;
	while (lineBegin < strlen (text)) {
		charCount = 0;
		buffCount = 0;
		do {
			buffer[buffCount] = text[lineBegin + charCount];
			buffer[buffCount+1] = 0;
			if (buffer[buffCount] == ' ') {
				lastSpace = 0;
			} else if (buffer[buffCount] == '\n') {
				lineCount++;
				charCount++;
				break;
			}
			lastSpace++;
			buffCount++;
			charCount++;
		} while (text[lineBegin + charCount] && (text_length (font, buffer) < region->w - 20));
		if ((lastSpace > 0) && (text_length (font, buffer) >= region->w - 20)) {
			charCount -= lastSpace - 1;
			buffer[buffCount - lastSpace] = 0;
		} else {
			buffer[buffCount] = 0;
		}
		textout (env->db, font, buffer, region->x + 5,
			region->y + (lineCount * text_height (font)) + 5, WHITE);
		lineBegin = lineBegin + charCount;
		charCount = 0;
		lineCount++;
	}
	env->make_update (region->x, region->y, region->w, region->h);
}

void draw_buystuff(GLOBALDATA *global, ENVIRONMENT *env, PLAYER *pl)
{

	env->make_update (0, 0, SCREEN_WIDTH, SCREEN_HEIGHT);
	show_mouse (NULL);
	
	draw_circlesBG (global, env->db, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT);
	draw_sprite (env->db, (BITMAP *) global->gfxData.M[DONE_IMAGE].dat, HALF_WIDTH - 100, SCREEN_HEIGHT - 50);
	draw_sprite (env->db, (BITMAP *) global->gfxData.M[FAST_UP_ARROW_IMAGE].dat, SCREEN_WIDTH - STUFF_BAR_WIDTH - 30, HALF_HEIGHT - 50);
	draw_sprite (env->db, (BITMAP *) global->gfxData.M[UP_ARROW_IMAGE].dat, SCREEN_WIDTH - STUFF_BAR_WIDTH - 30, HALF_HEIGHT - 25);
	draw_sprite (env->db, (BITMAP *) global->gfxData.M[DOWN_ARROW_IMAGE].dat, SCREEN_WIDTH - STUFF_BAR_WIDTH - 30, HALF_HEIGHT);
	draw_sprite (env->db, (BITMAP *) global->gfxData.M[FAST_DOWN_ARROW_IMAGE].dat, SCREEN_WIDTH - STUFF_BAR_WIDTH - 30, HALF_HEIGHT + 25);
	drawing_mode (DRAW_MODE_TRANS, NULL, 0, 0);

	for (z = 0; z < HALF_WIDTH - 200; z++) {
		set_trans_blender (0, 0, 0, (int)((double)((double)z / (HALF_WIDTH - 200)) * 240) + 15);
		vline (env->db, z, 0, 29, pl->color);
	}

	for (z = 0; z < HALF_WIDTH - 200; z++) {
		set_trans_blender (0, 0, 0, (int)((double)((double)z / (HALF_WIDTH - 200)) * 240) + 15);
		vline (env->db, (SCREEN_WIDTH-1) - z, 0, 29, pl->color);
	}

	solid_mode ();

}

int btps;
void draw_weapon_list(GLOBALDATA *global, ENVIRONMENT *env, PLAYER *pl, int scroll, int pressed)
{
	int slot;
	double tempbtps = (SCREEN_HEIGHT - 55) / STUFF_BAR_HEIGHT;
	// To be sure it rounds down
	btps = (int)tempbtps;
	if (tempbtps < btps)
		btps--;

	for (slot = 1, zzz = scroll; (slot < btps) && (zzz < env->numAvailable); zzz++) {
		int itemNum = env->availableItems[zzz];
		draw_sprite (env->db, (BITMAP *)global->gfxData.stuff_bar[(pressed == itemNum)?1:0], SCREEN_WIDTH - STUFF_BAR_WIDTH, slot * STUFF_BAR_HEIGHT);
		
		draw_sprite(env->db, (BITMAP *) global->gfxData.stuff_icon_base, SCREEN_WIDTH - STUFF_BAR_WIDTH, (slot * STUFF_BAR_HEIGHT));
		draw_sprite(env->db, (BITMAP *) global->gfxData.STOCK_IMAGE[itemNum].dat, SCREEN_WIDTH - STUFF_BAR_WIDTH, (slot * STUFF_BAR_HEIGHT) - 5);
		env->make_update (SCREEN_WIDTH - STUFF_BAR_WIDTH, slot * STUFF_BAR_HEIGHT, STUFF_BAR_WIDTH, STUFF_BAR_HEIGHT + 5);

		if (itemNum > WEAPONS - 1) {	/* Items part */

			textout (env->db, font,
				 item[itemNum - WEAPONS].name, SCREEN_WIDTH - STUFF_BAR_WIDTH + 45, (slot * STUFF_BAR_HEIGHT) + 5, BLACK);
			textprintf (env->db, font, SCREEN_WIDTH - STUFF_BAR_WIDTH + 45, (slot * STUFF_BAR_HEIGHT) + (STUFF_BAR_HEIGHT/2), BLACK, "Qty. in stock: %d", pl->ni[itemNum - WEAPONS]);
			sprintf (buf, "$%d", item[itemNum - WEAPONS].cost);
			textout (env->db, font, buf,
				 SCREEN_WIDTH - 45 - text_length (font, buf), (slot * STUFF_BAR_HEIGHT) + 5, BLACK);
			sprintf (buf, "for %d", item[itemNum - WEAPONS].amt);
			textout (env->db, font, buf,
				 SCREEN_WIDTH - 45 - text_length (font, buf), (slot * STUFF_BAR_HEIGHT) + (STUFF_BAR_HEIGHT/2), BLACK);
		} else {			/* Weapons part */

			textout (env->db, font, weapon[itemNum].name, SCREEN_WIDTH - STUFF_BAR_WIDTH + 45, (slot * STUFF_BAR_HEIGHT) + 5, BLACK);
			textprintf (env->db, font, SCREEN_WIDTH - STUFF_BAR_WIDTH + 45, (slot * STUFF_BAR_HEIGHT) + (STUFF_BAR_HEIGHT/2), BLACK, "Qty. in stock: %d", pl->nm[itemNum]);
			sprintf (buf, "$%d", weapon[itemNum].cost);
			textout (env->db, font, buf,
				 SCREEN_WIDTH - 45 - text_length (font, buf), (slot * STUFF_BAR_HEIGHT) + 5, BLACK);
			sprintf (buf, "for %d", weapon[itemNum].amt);
			textout (env->db, font, buf, SCREEN_WIDTH - 45 - text_length (font, buf), (slot * STUFF_BAR_HEIGHT) + (STUFF_BAR_HEIGHT/2), BLACK);
		}
		slot++;
	}

}

void buystuff (GLOBALDATA *global, ENVIRONMENT *env)
{
	int pl, done, updatew[THINGS], updatename, pressed, scroll, lb;
	int hoverOver = 0;
	char buf[50];

	global->updateCount = cclock = lb = env->mouseclock = 0;
	fi = global->stopwindow = updatename = scroll = 1;
	if (global->command != global->rounds) {
		for (z = 0; z < global->numPlayers; z++) {
			global->players[z]->money = (int) (global->players[z]->money * global->interest);
		}
	}

	env->generateAvailableItemsList ();

	for (pl = 0; pl < global->numPlayers; pl++) {
		if ((int)global->players[pl]->type != HUMAN_PLAYER) {
			int pressed = 0;
			while (global->players[pl]->money > 0 && pressed != -1) {
				pressed = global->players[pl]->chooseItemToBuy ();
				if (pressed == -1)
					break;
				if (pressed < WEAPONS && weapon[pressed].warhead)
					continue;
				if (global->players[pl]->money >= ((pressed > WEAPONS - 1) ? item[pressed - WEAPONS].cost : weapon[pressed].cost)
				    && ((pressed > WEAPONS - 1)
					? global->players[pl]->ni[pressed - WEAPONS] : global->players[pl]->nm[pressed]) < 99) {
					updatename = 1;
					if (pressed > WEAPONS - 1) {
						global->players[pl]->money -= item[pressed - WEAPONS].cost;
						global->players[pl]->ni[pressed - WEAPONS] += item[pressed - WEAPONS].amt;
						if (global->players[pl]->ni[pressed - WEAPONS] > 99)
							global->players[pl]->ni[pressed - WEAPONS] = 99;
					} else {
						global->players[pl]->money -= weapon[pressed].cost;
						global->players[pl]->nm[pressed] += weapon[pressed].amt;
						if (global->players[pl]->nm[pressed] > 99)
							global->players[pl]->nm[pressed] = 99;
					}
				}
			}
			continue;
		}

		done = 0;
		updatename = scroll = 1;
		pressed = -1;

		draw_buystuff (global, env, global->players[pl]);
		
		for (z = 0; z < THINGS; z++)
			updatew[z] = 1;
		while (!done) {

#ifdef LINUX
			usleep (1);

#endif /*  */
			while (cclock > 0) {
				cclock--;
				if (!lb && mouse_b & 1 && mouse_x >= HALF_WIDTH - 100 && mouse_x < HALF_WIDTH + 100 && mouse_y >= SCREEN_HEIGHT - 50 && mouse_y < SCREEN_HEIGHT - 25)
					done = 1;
				if (!lb && mouse_b & 1)
					env->mouseclock = 0;
				lb = (mouse_b & 1) ? 1 : 0;

				//Keyboard control
				if ((key[KEY_UP]) && (scroll > 1)
				    && (!env->mouseclock)) {
					scroll--;
					for (z = 0; z < THINGS; z++)
						updatew[z] = 1;
				}
				if ((key[KEY_PGUP]) && (scroll > 1)
				    && (!env->mouseclock)) {
					scroll -= btps / 2;
					if (scroll < 1)
						scroll = 1;
					for (z = 0; z < THINGS; z++)
						updatew[z] = 1;
				}
				if ((key[KEY_DOWN])
				    && (scroll <= env->numAvailable - btps)
				    && (!env->mouseclock)) {
					scroll++;
					for (z = 0; z < THINGS; z++)
						updatew[z] = 1;
				}
				if ((key[KEY_PGDN])
					&& (scroll <= env->numAvailable - btps + 1)
					&& (!env->mouseclock)) {
					scroll += btps / 2;
					if (scroll > env->numAvailable - btps + 1)
						scroll = env->numAvailable - btps + 1;
					for (z = 0; z < THINGS; z++)
						updatew[z] = 1;
				}

				//Mouse control
				if (mouse_x >= SCREEN_WIDTH - STUFF_BAR_WIDTH && mouse_x < SCREEN_WIDTH) {
					int newlyOver;
					zz = 0;
					for (z = 1, zzz = scroll; z < btps; z++, zzz++) {
						if (mouse_y >= z * STUFF_BAR_HEIGHT && mouse_y < (z * STUFF_BAR_HEIGHT) + 30) {
							zz = 1;
							break;
						}
					}
					if (zz)
						newlyOver = env->availableItems[zzz];
					else
						newlyOver = -1;
					if (hoverOver != newlyOver) {
						char description[1024];
						BOX area = {20, 60, 300, 400};

						if (newlyOver > -1) {
							if (newlyOver < WEAPONS) {
								WEAPON *weap = &weapon[newlyOver];
								sprintf (description, "Radius: %d\nYield: %d\n\n%s",
									weap->radius, weap->damage, weap->description);
							} else {
								ITEM *it = &item[newlyOver - WEAPONS];
								sprintf (description, "%s",
									it->description);
							}
						} else {
							description[0] = 0;
						}
						hoverOver = newlyOver;

						draw_text_in_box (env, &area, description);
					}
				}
				if (mouse_b & 1 && !env->mouseclock) {
					int scrollArrowPos = SCREEN_WIDTH - STUFF_BAR_WIDTH - 30;
					if (mouse_x >= scrollArrowPos && mouse_x < scrollArrowPos + 24) {
						if ((mouse_y >= HALF_HEIGHT - 50 && mouse_y < HALF_HEIGHT - 25) && (scroll > 1)) {
							scroll -= btps / 2;
							if (scroll < 1)
								scroll = 1;
							for (z = 0; z < THINGS; z++)
								updatew[z] = 1;
						}
						if ((mouse_y >= HALF_HEIGHT - 24 && mouse_y < HALF_HEIGHT) && (scroll > 1)) {
							scroll--;
							for (z = 0; z < THINGS; z++)
								updatew[z] = 1;
						}
						if ((mouse_y >= HALF_HEIGHT + 1 && mouse_y < HALF_HEIGHT + 25) && (scroll <= env->numAvailable - btps)) {
							scroll++;
							for (z = 0; z < THINGS; z++)
								updatew[z] = 1;
						}
						if ((mouse_y >= HALF_HEIGHT + 25 && mouse_y < HALF_HEIGHT + 50) && (scroll <= env->numAvailable - btps + 1)) {
							scroll += btps / 2;
							if (scroll > env->numAvailable - btps + 1)
								scroll = env->numAvailable - btps + 1;
							for (z = 0; z < THINGS; z++)
								updatew[z] = 1;
						}
					}
				}
				if (mouse_b & 1) {
					zz = 0;
					for (z = 1, zzz = scroll; z < btps; z++, zzz++) {
						if (mouse_x >= SCREEN_WIDTH - STUFF_BAR_WIDTH && mouse_x < SCREEN_WIDTH && mouse_y >= z * STUFF_BAR_HEIGHT && mouse_y < (z * STUFF_BAR_HEIGHT) + 30) {
							zz = 1;
							if (pressed > -1)
								updatew[pressed]
									= 1;
							pressed = env->availableItems[zzz];
							updatew[env->availableItems[zzz]] = 1;
						}
					}
					if (!zz) {
						if (pressed > -1)
							updatew[pressed] = 1;
						pressed = -1;
					}
				}
				if (pressed > -1 && !mouse_b & 1) {
					updatew[pressed] = 1;
					if (global->players[pl]->money >= ((pressed > WEAPONS - 1) ? item[pressed - WEAPONS].cost : weapon[pressed].cost)
					    && ((pressed > WEAPONS - 1)
						? global->players[pl]->ni[pressed - WEAPONS] : global->players[pl]->nm[pressed]) < 999) {
						updatename = 1;
						if (pressed > WEAPONS - 1) {
							global->players[pl]->money -= item[pressed - WEAPONS].cost;

							//original: tanks[pl].ni[pressed] -> I think it was mistake --v
							global->players[pl]->ni[pressed - WEAPONS] += item[pressed - WEAPONS].amt;
							if (global->players[pl]->ni[pressed - WEAPONS] > 999)
								global->players[pl]->ni[pressed - WEAPONS] = 999;
						} else {
							global->players[pl]->money -= weapon[pressed].cost;
							global->players[pl]->nm[pressed] += weapon[pressed].amt;
							if (global->players[pl]->nm[pressed] > 999)
								global->players[pl]->nm[pressed] = 999;
						}
					}
					pressed = -1;
				}
				env->mouseclock++;
				if (env->mouseclock > 5)
					env->mouseclock = 0;
			}
			show_mouse (NULL);
			if (fi) {
				for (z = 0; z < THINGS; z++)
					updatew[z] = 1;
			}
			if (updatename) {
				updatename = 0;
				env->make_update (HALF_WIDTH - 315, 0, 400, 30);
				draw_sprite (env->db, (BITMAP *) global->gfxData.stuff_bar[0], HALF_WIDTH - 200, 0);
				textprintf (env->db, font, HALF_WIDTH - 155, 5, BLACK, "Player %d: %s", pl + 1, global->players[pl]->getName ());
				textprintf (env->db, font, HALF_WIDTH - 155, 17, BLACK, "Money: $%d", global->players[pl]->money);
				sprintf (buf, "Round: %d/%d", (int)(global->rounds - global->command) + 1, (int)global->rounds);
				textout (env->db, font, buf, HALF_WIDTH + 170 - text_length (font, buf), 5, BLACK);
				sprintf (buf, "Score: %d", global->players[pl]->score);
				textout (env->db, font, buf, HALF_WIDTH + 155 - text_length (font, buf), 17, BLACK);
			}

			draw_weapon_list(global, env, global->players[pl], scroll, pressed);
			env->make_update (mouse_x, mouse_y, ((BITMAP *) (global->gfxData.M[0].dat))->w, ((BITMAP *) (global->gfxData.M[0].dat))->h);
			env->make_update (lx, ly, ((BITMAP *) (global->gfxData.M[0].dat))->w, ((BITMAP *) (global->gfxData.M[0].dat))->h);
			lx = mouse_x;
			ly = mouse_y;
			show_mouse (env->db);
			if (fi) {
				change (env->db);
				fi = 0;
			}

			else
				env->do_updates ();
		}
	}
	clear_keybuf ();
}

//Juraj Michalek -> convert this code back to C++ :-)))
float crand ()			// Pavel Minayev
{
	return 2.0 * rand () / RAND_MAX - 1;
}


struct moon {
	int radius;
	int x;
	int y;
	double lambda;
	int octaves;
	double smoothness;
	int col1;
	int col2;
};

double getv (int color)
{
	float h, s, v;
	int r, g, b;

	r = getr (color);
	g = getg (color);
	b = getb (color);
	rgb_to_hsv (r, g, b, &h, &s, &v);

	return (v);
}

void draw_moons (GLOBALDATA *global, ENVIRONMENT *env, int xoffset, int yoffset, int width, int heightx)
{
	int numMoons = rand () % 10;
	if (numMoons == 0)
		return;
	struct moon moons[numMoons];

	for (int count = 0; count < numMoons; count++) {
		moons[count].smoothness = (rand () % 20) + 3;
		moons[count].octaves = rand () % 4 + 6;
		moons[count].lambda = (rand () % 60 + 30) * (1.00/100);

		moons[count].radius = (int)((rand () % 100 / 200.0) * (rand () % 100 / 200.0) * SCREEN_WIDTH);
	 	moons[count].x = rand () % SCREEN_WIDTH;
	 	moons[count].y = rand () % SCREEN_HEIGHT;
		moons[count].col1 = makecol16 (rand () % 255, rand () % 255, rand () % 255);
		moons[count].col2 = makecol16 (rand () % 255, rand () % 255, rand () % 255);
	}

	drawing_mode (DRAW_MODE_TRANS, NULL, 0, 0);
	for (int x = 0; x < SCREEN_WIDTH; x++) {
		for (int y = 0; y < SCREEN_HEIGHT; y++) {
			double edgeval = 1;
			for (int count = 0; count < numMoons; count++) {
				int xpos = moons[count].x;
				int ypos = moons[count].y;
				int radius = moons[count].radius;
		
				int xdist = xpos - x;
				if (xdist > radius)
					continue;
				int ydist = ypos - y;
				if (ydist > radius)
					continue;
				double xval = (double)xdist / radius;
				double yval = (double)ydist / radius;
				double distance2 = (xdist * xdist) + (ydist * ydist);
				double distance = sqrt (distance2);

				int edgepixel = FALSE;
				if (distance <= (double)radius + 1) {
					if (distance >= (double)radius) {
						edgeval *= 1 - (distance - radius);
						edgepixel = TRUE;
					}
				}
				if (xval > 0 && distance <= (double)radius + 1) {
					double smoothness = moons[count].smoothness;
					double lambda = moons[count].lambda;
					int octaves = moons[count].octaves;

					double thetax = asin (xval) * 180 / PI;
					double thetay = acos (yval) * 180 / PI;
					double percentVal = 0;
					double offset = 0;

					offset += (perlin2DPoint (1.0, smoothness, xoffset + xpos + thetax, yoffset + ypos + thetay, lambda, octaves) + 1) / 2;
					percentVal += (perlin2DPoint (1.0, smoothness, xoffset + xpos * 1000 + thetax, yoffset + ypos * 1000 + thetay, lambda, octaves) + 1) / 2;

					set_add_blender (0, 0, 0, (int)(edgeval * xval * percentVal * offset * 255));
					putpixel (env->sky, x, y, moons[count].col1);
					set_add_blender (0, 0, 0, (int)(edgeval * xval * (1 - percentVal) * offset * 255));
					putpixel (env->sky, x, y, moons[count].col2);
				}
				if (edgepixel) {
					// How much of the moon behind should show
					edgeval = 1 - edgeval;
				}
				if (distance <= (double)radius)
					break;
			}
		}
	}
	solid_mode ();
}

void generate_sky (GLOBALDATA *global, ENVIRONMENT *env, int xoffset, int yoffset, int width, int heightx)
{
	//BITMAP *sky_gradient_strip = global->gfxData.sky_gradient_strips[global->cursky];
	double messiness = (rand () % 100 / 1000.0 + 0.05);

	for (int x = 0; x < SCREEN_WIDTH; x++) {
	        for (int y = 0; y < SCREEN_HEIGHT - MENU; y++) {
			double offset = 0;

			if (global->detailedSky)
				offset += perlin2DPoint (1.0, 200, xoffset + x, yoffset + y, 0.3, 6) * ((SCREEN_HEIGHT - MENU) * messiness);

			if (global->ditherGradients)
                        	offset += rand () % 10 - 5;

			while (y + offset < 0)
				offset /= 2;
			while (y + offset + 1 > (SCREEN_HEIGHT - MENU))
				offset /= 2;

			putpixel (env->sky, x, y,
				gradientColorPoint (sky_gradients[global->cursky], SCREEN_HEIGHT - MENU, y + offset));
			//putpixel (env->sky, x, y, getpixel (sky_gradient_strip, 0, y + offset));
		}
	}
	draw_moons (global, env, rand (), rand (), SCREEN_WIDTH, SCREEN_HEIGHT);
}

void generate_land (GLOBALDATA *global, ENVIRONMENT *env, int xoffset, int yoffset, int width, int heightx)
{
	const int land_height = heightx * 5 / 6;
	double	smoothness = 100;
	int	octaves = 8;
	double	lambda = 0.25;
	double	depthStrip[2][SCREEN_HEIGHT];

	int landType = (env->landType == LANDTYPE_RANDOM)? (rand () % LANDTYPE_PLAIN) + 1 : (int)env->landType;

	switch (landType) {
		case LANDTYPE_MOUNTAINS:
			smoothness = 200;
			octaves = 8;
			lambda = 0.65;
			break;
		case LANDTYPE_CANYONS:
			smoothness = 50;
			octaves = 8;
			lambda = 0.25;
			break;
		case LANDTYPE_HILLS:
			smoothness = 600;
			octaves = 6;
			lambda = 0.40;
			break;
		case LANDTYPE_FOOTHILLS:
			smoothness = 1200;
			octaves = 3;
			lambda = 0.25;
			break;
		case LANDTYPE_PLAIN:
			smoothness = 4000;
			octaves = 2;
			lambda = 0.2;
			break;
		default:
			break;
	}

	if (global->detailedLandscape)
		memset (depthStrip[1], 0, SCREEN_HEIGHT * sizeof (double));

	for (int x = 0; x < SCREEN_WIDTH; x++) {
		int depth = 0;
		height[x] = ((perlin2DPoint (1.0, smoothness, xoffset + x, yoffset, lambda, octaves) + 1) / 2);

		if (global->detailedLandscape) {
			memcpy (depthStrip[0], depthStrip[1], SCREEN_HEIGHT * sizeof(double));
			for (depth = 1; depth < SCREEN_HEIGHT; depth++) {
				depthStrip[1][depth] = ((perlin2DPoint (1.0, smoothness, xoffset + x, yoffset + depth, lambda, octaves) + 1) / 2 * land_height - (SCREEN_HEIGHT - depth));
				if (depthStrip[1][depth] > height[x] * land_height)
					depthStrip[1][depth] = height[x] * land_height;
			}
			depthStrip[1][0] = 0;
			depth = 1;
		}

		height[x] *= land_height;
		for (int y = 0; y <= height[x]; y++) {
			double offset = 0;
			int color = 0;
			double shade = 0;

			if (global->detailedLandscape) {
				double bot, top, minBot, maxTop, btdiff, i;
				double a1, a2, angle;
				while ((depthStrip[1][depth] <= y) && (depth < SCREEN_HEIGHT))
					depth++;
				bot = (depthStrip[0][depth - 1] + depthStrip[1][depth - 1]) / 2;
				top = (depthStrip[0][depth] + depthStrip[1][depth]) / 2;
				minBot = MIN (depthStrip[0][depth - 1], depthStrip[1][depth - 1]);
				maxTop = MAX (depthStrip[0][depth], depthStrip[1][depth]);
				btdiff = maxTop - minBot;
				i = (y - bot) / btdiff;
				a1 = atan2 (depthStrip[0][depth - 1] - depthStrip[1][depth - 1], 1.0) * 180 / PI + 180;
				a2 = atan2 (depthStrip[0][depth] - depthStrip[1][depth], 1.0) * 180 / PI + 180;

				angle = interpolate (a1, a2, i);
				shade = global->slope[(int)angle][0];
			}

			if (global->ditherGradients)
				offset += rand () % 10 - 5;

			if (global->detailedLandscape)
				offset += (SCREEN_HEIGHT - depth) * 0.5;

			while (y + offset < 0)
				offset /= 2;
			while (y + offset > SCREEN_HEIGHT)
				offset /= 2;

			color = gradientColorPoint (land_gradients[global->curland], SCREEN_HEIGHT, y + offset);
			//color = landtable[y + offset];
			if (global->detailedLandscape) {
				float h, s, v;
				int r, g, b;

				r = getr (color);
				g = getg (color);
				b = getb (color);
				rgb_to_hsv (r, g, b, &h, &s, &v);
				shade += (double)(rand () % 1000 - 500) * (1.0/10000);
				if (shade < 0)
					v += v * shade * 0.5;
				else
					v += (1 - v) * shade * 0.5;
				hsv_to_rgb (h, s, v, &r, &g, &b);
				color = makecol16 (r, g, b);
			}

			putpixel (env->terrain, x, SCREEN_HEIGHT - y, color);
		}
	}
}


BITMAP *create_gradient_strip (gradient *grad, int length)
{
	BITMAP *strip;
	int color;
	int currLine;

	strip = create_bitmap (1, length);
	clear_to_color (strip, BLACK);

	for (currLine = 0;currLine < length; currLine++) {
		color = gradientColorPoint (grad, length, currLine);
		putpixel (strip, 0, currLine, color);
	}

	return (strip);
}


void set_level_settings (GLOBALDATA *global, ENVIRONMENT *env)
{
	int taken[MAXPLAYERS];
	BITMAP *sky_gradient_strip, *land_gradient_strip;
	int z, peak_height = 0;
	int chosen = 0, chosenCount = 0;
	int objCount;
	TANK *ltank;
	int xoffset;

	srand (time (NULL));

	show_mouse (NULL);
	draw_sprite (screen, (BITMAP *) global->gfxData.M[1].dat, HALF_WIDTH - 120, HALF_HEIGHT + 115);
	textout_centre (screen, font, "Choosing colors", HALF_WIDTH, HALF_HEIGHT + 120, WHITE);
	// Choose appropriate gradients for sky and land
	while ((chosenCount < 60) && !chosen) {
		global->curland = rand () % LANDS;
		if (!global->gfxData.land_gradient_strips[global->curland])
			global->gfxData.land_gradient_strips[global->curland] = create_gradient_strip (land_gradients[global->curland], (SCREEN_HEIGHT - MENU));
		land_gradient_strip = global->gfxData.land_gradient_strips[global->curland];

		global->cursky = rand () % SKIES;
		if (!global->gfxData.sky_gradient_strips[global->cursky])
			global->gfxData.sky_gradient_strips[global->cursky] = create_gradient_strip (sky_gradients[global->cursky], (SCREEN_HEIGHT - MENU));
		sky_gradient_strip = global->gfxData.sky_gradient_strips[global->cursky];

		chosen = 1;
		for (z = 0; z < SCREEN_WIDTH; z++)
			if (peak_height < height[z])
				peak_height = (int)height[z];
		for (z = 0; z <= peak_height; z++)
		{
			int skyi, landi;
			double distance;
			skyi = getpixel (sky_gradient_strip, 0, (SCREEN_HEIGHT - MENU - z));
			landi = getpixel (land_gradient_strip, 0, z);

			distance = colorDistance (skyi, landi);
			if (distance < 30)
				chosen = 0;
		}
		chosenCount++;
	}

	show_mouse (NULL);
	draw_sprite (screen, (BITMAP *) global->gfxData.M[1].dat, HALF_WIDTH - 120, HALF_HEIGHT + 155);
	textout_centre (screen, font, "Rendering Sky", HALF_WIDTH, HALF_HEIGHT + 160, WHITE);
	xoffset = rand ();
	generate_sky (global, env, xoffset, 0, SCREEN_WIDTH, SCREEN_HEIGHT);

	show_mouse (NULL);
	draw_sprite (screen, (BITMAP *) global->gfxData.M[1].dat, HALF_WIDTH - 120, HALF_HEIGHT + 195);
	textout_centre (screen, font, "Rendering Landscape", HALF_WIDTH, HALF_HEIGHT + 200, WHITE);
	clear_to_color (env->terrain, PINK);

	//for (z = 0; z < (SCREEN_HEIGHT - MENU); z++) {
	//	landtable[z] = getpixel ((BITMAP *) land_gradient_strip, 0, z);
	//}
	xoffset = rand ();
	generate_land (global, env, xoffset, 0, SCREEN_WIDTH, SCREEN_HEIGHT);

	for (z = 0; z < global->numTanks; z++) {
		taken[z] = 0;
	}
	for (objCount = 0; (ltank = (TANK*)env->getNextOfClass (TANK_CLASS, &objCount)) && ltank; objCount++) {
		for (zz = rand () % global->numTanks; taken[zz]; zz = rand () % global->numTanks);
		taken[zz] = 1;
		ltank->x = (zz + 1) * (SCREEN_WIDTH / (global->numTanks + 1));
		ltank->y = (SCREEN_HEIGHT - (int)height[(int) ltank->x]) - (TANKHEIGHT - TANKSAG);
		ltank->newRound ();
	}
	for (z = 0; z < MAXPLAYERS * 3; z++)
		env->order[z] = NULL;
	global->maxNumTanks = global->numTanks;
	for (objCount = 0; (ltank = (TANK*)env->getNextOfClass (TANK_CLASS, &objCount)) && ltank; objCount++) {
		for (z = rand () % global->numTanks; env->order[z]; z = rand () % global->numTanks);
		env->order[z] = ltank;
	}
	if (global->turntype != TURN_RANDOM) {
		for (int index = 0; index < global->maxNumTanks - 1; index++) {
			int swap = FALSE;
			if (global->turntype == TURN_HIGH) {
				if (env->order[index]->player->score <
					env->order[index + 1]->player->score) {
					swap = TRUE;
				}
			} else if (global->turntype == TURN_LOW) {
				if (env->order[index]->player->score >
					env->order[index + 1]->player->score) {
					swap = TRUE;
				}
			}
			if (swap) {
				TANK *tempTank = env->order[index];
				env->order[index] =
					env->order[index + 1];
				env->order[index + 1] = tempTank;
				index = -1;
			}
		}
	}
}


void do_winner (GLOBALDATA *global, ENVIRONMENT *env)
{
	int winners[MAXPLAYERS], maxscore, length;
	char winnertext[MAXPLAYERS * 15];

	for (z = 0; z < MAXPLAYERS; z++)
		winners[z] = 0;
	for (z = 0; z < MAXPLAYERS * 15; z++)
		winnertext[z] = 0;
	maxscore = 0;
	for (z = 0; z < global->numPlayers; z++) {
		if (global->players[z]->score == maxscore)
			winners[z] = 1;
		if (global->players[z]->score > maxscore) {
			for (zz = 0; zz < global->numPlayers; zz++)
				winners[zz] = 0;
			winners[z] = 1;
			maxscore = global->players[z]->score;
		}
	}
	for (z = 0; z < global->numPlayers; z++) {
		if (winners[z]) {
			if (winnertext[0])
				strcat (winnertext, ", ");
			strcat (winnertext, global->players[z]->getName ());
		}
	}
	length = (text_length (font, winnertext) / 2) + 10;
	show_mouse (NULL);
	draw_circlesBG (global, env->db, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT);
	draw_sprite (env->db, (BITMAP *) global->gfxData.M[9].dat, HALF_WIDTH - 150, HALF_HEIGHT - 50);
	rectfill (env->db, HALF_WIDTH - length, HALF_HEIGHT + 60, HALF_WIDTH + length, HALF_HEIGHT + 75, WHITE);
	rect (env->db, HALF_WIDTH - length, HALF_HEIGHT + 60, HALF_WIDTH + length, HALF_HEIGHT + 75, BLACK);
	textout_centre (env->db, font, winnertext, HALF_WIDTH, HALF_HEIGHT + 64, BLACK);
	change (env->db);
	readkey ();
}


//draws indicaation bar
void graph_bar (ENVIRONMENT *env, int x, int y, long int col, int actual, int max)
{
	rect (env->db, x, y, x + max + 2, y + 8, BLACK);
	rectfill (env->db, x + 1, y + 1, x + 1 + actual, y + 7, col);
}


//draws indication bar - centred
void graph_bar_center (ENVIRONMENT *env, int x, int y, long int col, int actual, int max)
{
	rect (env->db, x, y, x + max + 2, y + 8, BLACK);
	rectfill (env->db, x + 1 + max / 2, y + 1, x + 1 + actual + max / 2, y + 7, col);
}


//Some global parametters
int ord;
void loadshields (GLOBALDATA *global, ENVIRONMENT *env)
{
	TANK *tank;
	int objCount;

	for (objCount = 0; (tank = (TANK*)env->getNextOfClass (TANK_CLASS, &objCount)) && tank; objCount++)
		tank->reactivate_shield ();
}

void change_wind_strength (ENVIRONMENT *env)
{
	if (env->windvariation == 0.0) {
		return;
	} else {
		env->wind = env->lastwind + (double)(rand () % (int)(env->windvariation * 100)) / 100 - (env->windvariation / 2);
		if (env->wind > (env->windstrength / 2))
		{
			env->wind = env->windstrength / 2;
		} else if (env->wind < (-env->windstrength / 2)) {
			env->wind = -env->windstrength / 2;
		}

		env->lastwind = env->wind;
	}
}

TANK *nextturn (GLOBALDATA *global, ENVIRONMENT *env)
{
	TANK *tank;
	int ordCurrently = ord;

	do {
		ord++;
		if (ord >= global->maxNumTanks)
			ord = 0;
	} while ((!env->order[ord]) && (ord != ordCurrently));
	tank = env->order[ord];
	global->currTank = tank;

	//Wind is blowing :-)
	change_wind_strength (env);

	if (tank) {
		tank->reactivate_shield ();
	}

	return tank;
}

void showRoundEndScoresAt (GLOBALDATA *global, ENVIRONMENT *env, BITMAP *bitmap, int x, int y, int winner)
{
	env->make_update (x - 100, y - 100, 201, 201);
	rectfill (bitmap, x - 100, y - 100, x + 100, y + 100, BLACK);
	rect (bitmap, x - 100, y - 100, x + 100, y + 100, WHITE);
	if (winner == -2) {
		textout_centre (bitmap, font, "Draw", x, y - 90, WHITE);
	} else {
		textprintf_centre (bitmap, font, x, y - 90, global->players[winner]->color, "Winner: %s", global->players[winner]->getName ());
	}
	textout_centre (bitmap, font, "Player Scores", x, y - 70, WHITE);
	for (z = 0; z < global->numPlayers; z++) {
		textprintf (bitmap, font, x - 90, (z * 10) + y - 50, global->players[z]->color, "%s:", global->players[z]->getName ());
		textprintf (bitmap, font, x + 10, (z * 10) + y - 50, WHITE, "%d", global->players[z]->score);
	}
}

int setSlideColumnDimensions (GLOBALDATA *global, ENVIRONMENT *env, int x, char reset)
{
	int pixelHeight;
	char	*done	= env->done;
	int	*dropTo = env->dropTo;
	int	*h	= env->h;
	int	*fp	= env->fp;
	double	*velocity = env->velocity;
	double	*dropIncr = env->dropIncr;

	if (x < 0 || x > (SCREEN_WIDTH-1)) {
		return (0);
	}

	if (reset) {
		h[x] = 0;
		fp[x] = 0;
		dropTo[x] = SCREEN_HEIGHT - 1;
	}
	done[x] = 0;

	// Calc the top and bottom of the column to slide

	// Find top-most non-PINK pixel
	for (pixelHeight = h[x]; pixelHeight < dropTo[x]; pixelHeight++)
		if (getpixel (env->terrain, x, pixelHeight) != PINK)
			break;
	h[x] = pixelHeight;

	// Find bottom-most PINK pixel
	for (pixelHeight = dropTo[x]; pixelHeight > h[x]; pixelHeight--)
		if (getpixel (env->terrain, x, pixelHeight) == PINK)
			break;
	dropTo[x] = pixelHeight;

	// Find bottom-most unsupported pixel
	for (; pixelHeight >= h[x]; pixelHeight--)
		if (getpixel (env->terrain, x, pixelHeight) != PINK)
			break;

	// If there's some processing to do
	if ((pixelHeight >= h[x]) && (h[x] < dropTo[x]))
	{
		fp[x] = pixelHeight - (int)h[x] + 1;
		return (0);
	} else {
		if (velocity[x])
			play_sample ((SAMPLE *) global->SOUND[10].dat, (int)((velocity[x] / 10) * 255), (int)((double)(x - HALF_WIDTH) / HALF_WIDTH * 128 + 128), 1000 - (int)((double)fp[x] / SCREEN_HEIGHT) * 1000, 0);
		h[x] = 0;
		fp[x] = 0;
		done[x] = 1;
		velocity[x] = 0;
		dropIncr[x] = 0;
		dropTo[x] = SCREEN_HEIGHT - 1;
		return (1);
	}
	return (0);
}


int drawFracture (GLOBALDATA *global, ENVIRONMENT *env, BITMAP *dest, BOX *updateArea, int x, int y, int angle, int width, int segmentLength, int maxRecurse, int recurseDepth)
{
	int branchCount;
	int x1, x2, x3;
	int y1, y2, y3;

	x1 = (int)(x + global->slope[angle][0] * width);
	y1 = (int)(y + global->slope[angle][1] * width);
	x2 = (int)(x - global->slope[angle][0] * width);
	y2 = (int)(y - global->slope[angle][1] * width);
	x3 = (int)(x + global->slope[angle][1] * segmentLength);
	y3 = (int)(y + global->slope[angle][0] * segmentLength);
	triangle (dest, x1, y1, x2, y2, x3, y3, PINK);

	if (recurseDepth == 0) {
		updateArea->x = x1;
		updateArea->y = y1;
		updateArea->w = x1;
		updateArea->h = y1;
	}
	updateArea->x = MIN (MIN (MIN (x1, x2), x3), updateArea->x);
	updateArea->y = MIN (MIN (MIN (y1, y2), y3), updateArea->y);
	updateArea->w = MAX (MAX (MAX (x1, x2), x3), updateArea->w);
	updateArea->h = MAX (MAX (MAX (y1, y2), y3), updateArea->h);

	if (recurseDepth < maxRecurse)
	{
		for (branchCount = 0; branchCount < 3; branchCount++)
		{
			if ((branchCount == 0) || (Noise (x + y + branchCount) < 0))
			{
				int newAngle, reduction;
				newAngle = (angle + (int)(Noise (x + y + 4) * 30)) % 360;
				reduction = 2;
				if (branchCount == 1) {
					newAngle = (int)(angle + 90 +
						(Noise (x + y + 25 + branchCount) * 22.5)) % 360;
					reduction = abs ((int)Noise (x + y + 1 + branchCount) * 4 + 3);
				} else if (branchCount == 2) {
					newAngle = (int)(angle + 270 +
						(Noise (x + y + 32 + branchCount) * 22.5)) % 360;
					reduction = abs ((int)Noise (x + y + 2 + branchCount) * 4 + 3);
				}
				drawFracture (global, env, dest, updateArea, x3, y3, newAngle, width / reduction, segmentLength / reduction, maxRecurse, recurseDepth + 1);
			}
		}
	}

	// Calculate width and height, previously right and bottom
	if (recurseDepth == 0) {
		updateArea->w -= updateArea->x;
		updateArea->h -= updateArea->y;
	}

	return (0);
}

int slideLand (GLOBALDATA *global, ENVIRONMENT *env)
{
	char	*done	= env->done;
	int	*dropTo = env->dropTo;
	int	*h	= env->h;
	int	*fp	= env->fp;
	double	*velocity = env->velocity;
	double	*dropIncr = env->dropIncr;

	// land-slide, make it fall etc.
	int allDone = 1;
	if (env->landSlideType == LANDSLIDE_NONE)
		return (allDone);

	for (zz = 0; zz < SCREEN_WIDTH; zz++) {
		if (!done[zz]) {
			allDone = 0;
			if (env->landSlideType == LANDSLIDE_GRAVITY) {
				if (fp[zz] > 0) {
					velocity[zz] += env->gravity;
					dropIncr[zz] += velocity[zz];
					if (dropIncr[zz] >= 1) {
						if (dropIncr[zz] > dropTo[zz] - (h[zz] + fp[zz])) {
							dropIncr[zz] = dropTo[zz] - (h[zz] + fp[zz]) + 1;
						}
						blit (env->terrain, env->terrain, zz, h[zz] - (int)dropIncr[zz], zz, h[zz], 1, fp[zz] + (int)dropIncr[zz]);
						env->make_bgupdate (zz, h[zz] - (int)dropIncr[zz], 1, fp[zz] + ((int)dropIncr[zz] * 2) + 1);
						env->make_update (zz, h[zz] - (int)dropIncr[zz], 1, fp[zz] + ((int)dropIncr[zz] * 2) + 1);
						h[zz] += (int)dropIncr[zz];
						dropIncr[zz] -= (int)dropIncr[zz];
					}
					setSlideColumnDimensions (global, env, zz, FALSE);
				} else {
					setSlideColumnDimensions (global, env, zz, FALSE);
				}
			} else if (env->landSlideType == LANDSLIDE_INSTANT) {
				if (fp[zz] > 0) {
					env->make_bgupdate (zz, h[zz], 1, dropTo[zz] - h[zz] + 1);
					env->make_update (zz, h[zz], 1, dropTo[zz] - h[zz] + 1);
					done[zz] = 1;
					blit (env->terrain, env->terrain, zz, h[zz], zz, dropTo[zz] - fp[zz] + 1, 1, fp[zz]);
					vline (env->terrain, zz, h[zz], dropTo[zz] - fp[zz], PINK);
				}
				setSlideColumnDimensions (global, env, zz, FALSE);
			}
		}
	}

	return (allDone);
}

void drawTopBar (GLOBALDATA *global, ENVIRONMENT *env, BITMAP *dest)
{
	TANK *tank = global->currTank;
	char *name = "";
	int color = 0;
	int wind_col1 = 0, wind_col2 = 0;

	if (tank) {
		name = global->currTank->player->getName ();
		color = global->currTank->player->color;
	}

	global->updateMenu = 0;
	blit (global->gfxData.topbar, dest, 0, 0, 0, 0, SCREEN_WIDTH, MENU);

	if (tank) {
		textout (dest, font, name, 2, 2, BLACK);
		textout (dest, font, name, 1, 1, color);
		textprintf (dest, font, 1, 11, BLACK, "Angle:");
		graph_bar_center (env, 50, 11, color, -(tank->a - 180) / 2, 180 / 2);
		// 0 is directly left, 180 points directly right
		textprintf (dest, font, 150, 11, BLACK, "%d", 180 - (tank->a - 90));

		textprintf (dest, font, 1, 21, BLACK, "Power:");
		graph_bar (env, 50, 20, color, (tank->p) / (MAX_POWER/90), 90);
		textprintf (dest, font, 150, 21, BLACK, "%d", tank->p);
		if (tank->cw < WEAPONS) {
			textprintf (dest, font, 100, 1, BLACK, "%s: %d",
				    weapon[tank->cw].name, tank->player->nm[tank->cw]);
//			textprintf (dest, font, 150, 11, BLACK, "Radius: %d", weapon[tank->cw].radius);
//			textprintf (dest, font, 150, 21, BLACK, "Damage: %d", weapon[tank->cw].damage);
		} else {
			textprintf (dest, font, 100, 1, BLACK, "%s: %d",
				    item[tank->cw - WEAPONS].name, tank->player->ni[tank->cw - WEAPONS]);
		}
		textprintf (dest, font, 350, 1, BLACK, "$%d", tank->player->money);
/*		if (tank->sh > 0 || tank->fs) {
			sprintf (buf, item[tank->sht].name);
			strcat (buf, ": %d");
			textprintf (dest, font, 350, 11, BLACK, buf, tank->player->ni[tank->sht]);
			textprintf (dest, font, 350, 11, BLACK, "Shield: %d", tank->sh);
		} else {
			textout (dest, font, "No Shield", 350, 11, BLACK);
		}*/
	}

	textprintf (dest, font, 500, 1, BLACK, "Round %d/%d", (int)(global->rounds - global->command) + 1, (int)global->rounds);
	textprintf (dest, font, 500, 11, BLACK, "Wind: ");
	if (env->wind > 0) {
		wind_col1 = 1;
		wind_col2 = 0;
	}
	if (env->wind < 0) {
		wind_col1 = 0;
		wind_col2 = 1;
	}
	rect (dest, 540, 12, (int)(540 + env->windstrength * 4 + 2), 18, BLACK);
	rectfill (dest, (int)(541 + env->windstrength * 2), 13,
		  (int) (541 + env->wind * 4 + env->windstrength * 2), 17,
		  makecol16 (200 * wind_col1, 200 * wind_col2, 0));
	global->stopwindow = 1;
	env->make_update (0, 0, SCREEN_WIDTH, MENU);
	global->stopwindow = 0;
}

void doNaturals (GLOBALDATA *global, ENVIRONMENT *env)
{
	if (env->meteors) {
		int chance = (int)(600 / env->meteors) + 100;
		if (!(rand () % chance)) {
			MISSILE *newmis;
			int ca = ((rand () % 160) + (360 - 80)) % 360;
			double mxv = global->slope[ca][0] * 5;
			double myv = global->slope[ca][1] * 5;

			newmis = new MISSILE(global, env,
				rand () % SCREEN_WIDTH, 0,
				mxv, myv, SML_METEOR + (rand () % (int)env->meteors));
			if (!newmis) {
				perror ("atanks.cc: Failed allocating memory for newmis in doNaturals");
				exit (1);
			}
			newmis->player = NULL;
		}
	}
	if (env->lightning) {
		int chance = (int)(600 / env->lightning) + 100;
		if (!(rand () % chance)) {
			BEAM *newbeam;
			int ca = ((rand () % 160) + (360 - 80)) % 360;
			//double mxv = global->slope[ca][0] * 400;
			//double myv = global->slope[ca][1] * 400;

			newbeam = new BEAM (global, env,
				rand () % SCREEN_WIDTH, 0,
				ca, SML_LIGHTNING + (rand () % (int)env->lightning));
			if (!newbeam) {
				perror ("atanks.cc: Failed allocating memory for newbeam in doNaturals");
				exit (1);
			}
			newbeam->player = NULL;
		}
	}
}

void game (GLOBALDATA *global, ENVIRONMENT *env)
{
	int tanksfall;
	int tanklife, tlt, dclock;
	int lb, ca;
	int allDone, anyExploding;
	int roundEndCount = 0;
	int nextTankSelected = FALSE;
	int humanPlayers = 0;
	int computerPlayersOnly = FALSE;
	int skippingComputerPlay = FALSE;

	TANK **tank, *ltank;
	MISSILE *missile;
	TELEPORT *teleport;
	DECOR *decor;
	BEAM *beam;
	EXPLOSION *explosion;
	FLOATTEXT *floattext;

	int bCount;

	int objCount, count;

	tank = &global->currTank;
		
	for (int doneCount = 0; doneCount < SCREEN_WIDTH; doneCount++)
		env->done[doneCount] = 1;
	
	for (count  = 0; count < global->numPlayers; count++)
		global->players[count]->newRound ();
	buystuff (global, env);
	set_level_settings (global, env);
	for (objCount = 0; (ltank = (TANK*)env->getNextOfClass (TANK_CLASS, &objCount)) && ltank; objCount++) {
		ltank->newRound ();
		if ((int)ltank->player->type == HUMAN_PLAYER) {
			humanPlayers++;
		}
	}
	if (!humanPlayers) {
		computerPlayersOnly = TRUE;
	}
	cclock = lx = ly = env->stage = tanksfall = 0;
	tlt = global->updateCount = dclock = global->stopwindow = 0;
	env->realm = env->am = 0;
	ca = 0;
	lb = env->mouseclock = env->pclock = 0;
	ord = 0;

	*tank = env->order[0];
	for (objCount = 0; (ltank = (TANK*)env->getNextOfClass (TANK_CLASS, &objCount)) && ltank; objCount++) {
		ltank->flashdamage = 0;
		ltank->boost_up_shield ();
	}
	winner = tanklife = -1;
	fi = global->updateMenu = 1;
	global->window.x = 0;
	global->window.y = 0;
	global->window.w = (SCREEN_WIDTH-1);
	global->window.h = (SCREEN_HEIGHT-1);
	bCount = 0;
	env->wind = (float)(rand () % (int)env->windstrength) - (env->windstrength / 2);
	env->lastwind = env->wind;
	while (1) {

#ifdef LINUX
		usleep (1);

#endif /*  */
		while (cclock > 0 || skippingComputerPlay) {
			cclock--;
			if (!lb && mouse_b & 1)
				env->mouseclock = 0;
			lb = (mouse_b & 1) ? 1 : 0;
			bCount += 360/32;
			z4 = 0;
			nextTankSelected = FALSE;
			anyExploding = 0;
			for (objCount = 0; (explosion = (EXPLOSION*)env->getNextOfClass (EXPLOSION_CLASS, &objCount)) && explosion; objCount++) {
				anyExploding++;
				explosion->explode ();
				explosion->applyPhysics ();
				if (explosion->destroy) {
					explosion->requireUpdate ();
					explosion->update ();
					delete explosion;
				}
			}
			for (objCount = 0; (teleport = (TELEPORT*)env->getNextOfClass (TELEPORT_CLASS, &objCount)) && teleport; objCount++) {
				teleport->applyPhysics ();
				if (teleport->destroy) {
					teleport->requireUpdate ();
					teleport->update ();
					delete teleport;
				}
			}
			for (objCount = 0; (decor = (DECOR*)env->getNextOfClass (DECOR_CLASS, &objCount)) && decor; objCount++) {
				decor->applyPhysics ();
				if (decor->destroy) {
					decor->requireUpdate ();
					decor->update ();
					delete decor;
				}
			}
			env->am = 0;
			for (objCount = 0; (missile = (MISSILE*)env->getNextOfClass (MISSILE_CLASS, &objCount)) && missile; objCount++) {
				env->am++;
				missile->hitSomething = 0;
				missile->applyPhysics ();
				missile->triggerTest ();
				if (missile->destroy) {
					missile->requireUpdate ();
					missile->update ();
					delete missile;
					tanksfall = 1;
				}
			}
			for (objCount = 0; (beam = (BEAM*)env->getNextOfClass (BEAM_CLASS, &objCount)) && beam; objCount++) {
				beam->applyPhysics ();
				if (beam->destroy) {
					beam->requireUpdate ();
					beam->update ();
					delete beam;
					tanksfall = 1;
				}
			}
			for (objCount = 0; (floattext = (FLOATTEXT*)env->getNextOfClass (FLOATTEXT_CLASS, &objCount)) && floattext; objCount++) {
				floattext->applyPhysics ();
				if (floattext->destroy) {
					floattext->requireUpdate ();
					floattext->update ();
					delete floattext;
				}
			}
			if (!anyExploding)
				doNaturals (global, env);
			allDone = slideLand (global, env);
			if (tanksfall && env->stage != 3) {
				for (objCount = 0; (ltank = (TANK*)env->getNextOfClass (TANK_CLASS, &objCount)) && ltank; count--, objCount++) {
					ltank->pen = 0;
					ltank->applyPhysics ();
					if (ltank->l <= 0 && !anyExploding) {
						ltank->explode ();
						if (ltank->creditTo) {
							if (ltank->player != ltank->creditTo) {	//enemy destroyed
								ltank->creditTo->money += (int)global->scoreUnitDestroyBonus;
							} else {	//self destroy - ugh foolish one :))
								ltank->creditTo->money -= (int)global->scoreUnitSelfDestroy;
							}
							ltank->creditTo = NULL;
						}
						if (ltank->destroy) {
							if ((int)ltank->player->type == HUMAN_PLAYER) {
								humanPlayers--;
								if (((int)global->skipComputerPlay) && (!humanPlayers) && (!computerPlayersOnly)) {
									skippingComputerPlay = TRUE;
									draw_sprite (env->db, (BITMAP *) global->gfxData.M[1].dat, HALF_WIDTH - 120, HALF_HEIGHT + 155);
									textout_centre (env->db, font, "Accelerated AI", HALF_WIDTH, HALF_HEIGHT + 160, WHITE);
									draw_sprite (screen, (BITMAP *) global->gfxData.M[1].dat, HALF_WIDTH - 120, HALF_HEIGHT + 155);
									textout_centre (screen, font, "Accelerated AI", HALF_WIDTH, HALF_HEIGHT + 160, WHITE);
								}
							}
							delete ltank;
							if (!*tank) {
								*tank = nextturn (global, env);
								if (*tank)
									(*tank)->fs = 0;
								nextTankSelected = TRUE;
							}
							if (global->numTanks <= 1) {
								skippingComputerPlay = FALSE;
								cclock = 0;
								env->stage = 3;
								global->currTank = NULL;
								fi = 1;
								global->window.x = 0;
								global->window.y = 0;
								global->window.w = (SCREEN_WIDTH-1);
								global->window.h = (SCREEN_HEIGHT-1);
								winner = -2;
								if (global->numTanks > 0) {
									for (z = 0; z < global->numPlayers; z++) {
										if (global->players[z]->tank)
											winner = z;
									}
								}
								if (winner >= 0) {
									global->players[winner]->score++;
									global->players[winner]->won++;
									global->players[winner]->money += (int)global->scoreRoundWinBonus;
								}
								bCount = 0;
								global->updateMenu = 1;
								for (z = 0; z < global->numPlayers; z++)
									global->players[z]->played++;
							}
						}
					}
				}
			}
			if (env->stage == 1) {
				if ((env->am == 0) && allDone && !anyExploding) {
					tanksfall = 0;
					env->stage = 2;
				}
			}
			if (env->stage == 2) {
				zz = 0;

				for (objCount = 0; (ltank = (TANK*)env->getNextOfClass (TANK_CLASS, &objCount)) && ltank; count--, objCount++) {
					zz += ltank->applyPhysics ();
				}
				if (zz ==  global->numTanks) {
					tanksfall = 1;
				}
				zz = 0;
				if (tanksfall) {
					for (objCount = 0; (ltank = (TANK*)env->getNextOfClass (TANK_CLASS, &objCount)) && ltank; objCount++) {
						ltank->pen = 0;
						if (ltank->l > 0)
							zz++;
					}
				}
				if (zz == global->numTanks) {
					env->stage = 0;
					winner = -2;
					for (z = 0; z < global->numPlayers; z++) {
						if (global->players[z]->tank && winner >= 0)
							winner = -1;
						if (global->players[z]->tank && winner == -2)
							winner = z;
					}
					if (winner >= 0) {
						global->players[winner]->score++;
					}
					if (winner == -1) {
						if (!nextTankSelected) {
							*tank = nextturn (global, env);
							(*tank)->fs = 0;
						}
					}
					if (winner >= 0 || winner == -2) {
						env->stage = 3;
						global->currTank = NULL;
						fi = 1;
						global->window.x = 0;
						global->window.y = 0;
						global->window.w = (SCREEN_WIDTH-1);
						global->window.h = (SCREEN_HEIGHT-1);
					}
					bCount = 0;
					global->updateMenu = 1;
				}
			}
			dclock++;
			if (dclock > 2) {
				dclock = 0;
				for (objCount = 0; (ltank = (TANK*)env->getNextOfClass (TANK_CLASS, &objCount)) && ltank; objCount++) {
					if (ltank->flashdamage) {
						if (ltank->flashdamage > 25 || ltank->l < 1) {
							ltank->damage = 0;
							ltank->flashdamage = 0;
							ltank->requireUpdate ();
						}
					}
				}
			}
			env->pclock++;
			if (env->pclock > 10)
				env->pclock = 0;
			if (*tank) {
				if ((*tank)->player->controlTank (*tank) == -1)
					return;
			} else if ((keypressed () || mouse_b) && !fi) {
				if (keypressed ())
					k = readkey ();
				else
					k = 0;
				if ((env->stage == 3) && (mouse_b || k >> 8 == KEY_ENTER || k >> 8 == KEY_ESC || k >> 8 == KEY_SPACE))
					return;
			}
			env->mouseclock++;
			if (env->mouseclock > 10)
				env->mouseclock = 0;
		}
		frames++;
		global->stopwindow = 1;
		env->make_update (mouse_x, mouse_y, ((BITMAP *) (global->gfxData.M[0].dat))->w, ((BITMAP *) (global->gfxData.M[0].dat))->h);
		env->make_update (lx, ly, ((BITMAP *) (global->gfxData.M[0].dat))->w, ((BITMAP *) (global->gfxData.M[0].dat))->h);
		global->stopwindow = 0;
		lx = mouse_x;
		ly = mouse_y;
		set_clip (env->db, 0, 0, (SCREEN_WIDTH-1), (SCREEN_HEIGHT-1));
		show_mouse (NULL);
		if (global->updateMenu) {
			drawTopBar (global, env, env->db);
		}
		set_clip (env->db, 0, MENU, (SCREEN_WIDTH-1), (SCREEN_HEIGHT-1));
		if (fi) {
			blit (env->sky, env->db, global->window.x, global->window.y - MENU, global->window.x, global->window.y, (global->window.w - global->window.x) + 1, (global->window.h - global->window.y) + 1);
			masked_blit (env->terrain, env->db, global->window.x, global->window.y, global->window.x, global->window.y, (global->window.w - global->window.x) + 1, (global->window.h - global->window.y) + 2);
		} else {
			env->replaceCanvas ();
		}

		for (objCount = 0, count = 0; (ltank = (TANK*)env->getNextOfClass (TANK_CLASS, &objCount)) && ltank; count++, objCount++) {
			if (env->stage < 3) {
				if (*tank == ltank) {
					ltank->draw (env->db, (int)(global->slope[bCount % 360][0] * 4));
					ltank->requireUpdate ();
				} else {
					ltank->draw (env->db, 0);
				}
				ltank->update ();
			}
			ltank->framelyAccounting ();
		}
		for (objCount = 0; (missile = (MISSILE*)env->getNextOfClass (MISSILE_CLASS, &objCount)) && missile; objCount++) {
			missile->draw (env->db);
			missile->update ();
		}
		for (objCount = 0; (beam = (BEAM*)env->getNextOfClass (BEAM_CLASS, &objCount)) && beam; objCount++) {
			beam->draw (env->db);
			beam->update ();
		}
		for (objCount = 0; (explosion = (EXPLOSION*)env->getNextOfClass (EXPLOSION_CLASS, &objCount)) && explosion; objCount++) {
			explosion->draw (env->db);
			explosion->update ();
		}
		for (objCount = 0; (teleport = (TELEPORT*)env->getNextOfClass (TELEPORT_CLASS, &objCount)) && teleport; objCount++) {
			if (teleport->object)
				teleport->draw (env->db);
			teleport->update ();
		}
		for (objCount = 0; (decor = (DECOR*)env->getNextOfClass (DECOR_CLASS, &objCount)) && decor; objCount++) {
			decor->draw (env->db);
			decor->update ();
		}
		for (objCount = 0; (floattext = (FLOATTEXT*)env->getNextOfClass (FLOATTEXT_CLASS, &objCount)) && floattext; objCount++) {
			floattext->draw (env->db);
			floattext->requireUpdate ();
			floattext->update ();
		}
		if (env->stage == 3) {
			roundEndCount++;
			if (roundEndCount > 60)
				showRoundEndScoresAt (global, env, env->db, SCREEN_WIDTH/2, SCREEN_HEIGHT/2, winner);
		}
		set_clip (env->db, 0, 0, (SCREEN_WIDTH-1), (SCREEN_HEIGHT-1));
		show_mouse (env->db);
		if (fi) {
			while (keypressed ()) {
				readkey ();
			}
			fi = 0;
			if (env->fog) {
				clear_to_color (screen, makecol16 (128,128,128));
			} else {
				quickChange (env->db);
			}
		} else {
			env->do_updates ();
			global->window.x = SCREEN_WIDTH;
			global->window.y = SCREEN_HEIGHT;
			global->window.w = -1;
			global->window.h = -1;
		}
	}
}

void print_text_help()
{
	cout	<< "-h\tThis screen" << endl
		<< "-fs\tFull screen" << endl;
}

void print_text_initmsg()
{
	printf ("Atomic Tanks Version %s (-h for help)\n", VERSION);
	printf ("Authors: \tTom Hudson (rewrite, additions, improvements)\n");
	printf ("\t\tStevante Software (original design)\n");
	printf ("\t\tKota543 Software (fixes and updates)\n");
	putchar ('\n');
}

void endgame_cleanup (GLOBALDATA *global, ENVIRONMENT *env)
{
	while (global->numPlayers > 0) {
		if (global->players[0]->tank)
			delete global->players[0]->tank;
		global->players[0]->tank = NULL;
		global->removePlayer (global->players[0]);
	}
	for (int objCount = 0; objCount < MAX_OBJECTS; objCount++) {
		if (env->objects[objCount]) {
			delete env->objects[objCount];
			env->objects[objCount] = NULL;
		}
	}
}

int main (int argc, char **argv)
{
	int signal;
	int stop = 0;
	string tmp;
	ENVIRONMENT *env;
	GLOBALDATA *global;
	FILE *configFile;
	char *homeDir;
	char fullPath[2048];

	global = new GLOBALDATA ();
	if (!global) {
		perror ("Allocating global");
		exit (1);
	}

	print_text_initmsg();
	
	if(argc == 2) {		/* Parse command-line switches */
		tmp = argv[1];
		
		if(tmp == SWITCH_HELP) {
			print_text_help();
			return 0;
		}

		if(tmp == SWITCH_FULL_SCREEN) {
			screen_mode = GFX_AUTODETECT_FULLSCREEN;
		}

		if(tmp == SWITCH_WINDOWED) {
			screen_mode = GFX_AUTODETECT_WINDOWED;
		}
	}
	
	env = init_game_settings (global);

	homeDir = getenv ("HOME");
	if (homeDir) {
		sprintf (fullPath ,"%s/%s", homeDir, ".atanks-config");
	} else {
		sprintf (fullPath, "%s", ".atanks-config");
	}
	
	configFile = fopen (fullPath, "rb");
	if (configFile) {
		int key;
		fread (&key, sizeof (int), 1, configFile);
		if (key == GFILE_KEY) {
			global->loadFromFile (configFile);
			env->loadFromFile (configFile);
			if (loadPlayers (global, env, configFile)) {
				fclose (configFile);
				perror ("loadPlayers");
				exit (1);
			}
			fclose (configFile);
		} else {
			fprintf (stderr, "Invalid or corrupt config file \"%s\"\n", fullPath);
			fclose (configFile);
			exit (1);
		}
	} else {
		char *defaultNames[] = {
			"Bri",
			"Sar",
			"Bill",
			"Rog",
			"TY",
			"Wembly",
			"Kat",
			"Snork",
			"Kel",
			"MRK"
		};
		PLAYER *tempPlayer;

		tempPlayer = global->createNewPlayer (env);
		tempPlayer->setName ("Your Name");
		options (global, env, (MENUDESC*)tempPlayer->menudesc);
		for (int count = 0; count < 10; count++) {
			tempPlayer = global->createNewPlayer (env);
			tempPlayer->type = rand () % (LAST_PLAYER_TYPE - 1) + 1;
			tempPlayer->setName (defaultNames[count]);
		}
	}

	title (global);

	do {
		global->command = GLOBAL_COMMAND_MENU;
		signal = menu (global, env);

		if ((signal == SIG_QUIT_GAME) || 
			(global->command == GLOBAL_COMMAND_QUIT )){
			stop = 1;
		}
		
		if (global->command == GLOBAL_COMMAND_HELP) {
			instructions (global, env);
			continue;
		}

		if (global->command == GLOBAL_COMMAND_OPTIONS) {
			options (global, env, NULL);
			continue;
		}

		if (global->command == GLOBAL_COMMAND_PLAYERS) {
			int editRetVal = 0;
			while (editRetVal >> 8 != KEY_ESC)
				editRetVal = editPlayers (global, env);
			continue;
		}

		if (global->command == GLOBAL_COMMAND_CREDITS) {
			credits (global, env);
			continue;
		}
		
		if(!stop) {
			if (selectPlayers (global, env) >> 8 == KEY_ESC)
				continue;

			newgame (global, env); 

			for (global->command = (int)global->rounds; global->command > 0; global->command--) {
				game (global, env);
			}

			if (global->command == 0)
				do_winner (global, env);

			endgame_cleanup (global, env);
		}

		if (global->command == -1) {
			stop = 1;
		}
		
		
	} while (!stop);

	configFile = fopen (fullPath, "wb");
	if (configFile) {
		int key = GFILE_KEY;

		fwrite (&key, sizeof (int), 1, configFile);
		global->saveToFile (configFile);
		env->saveToFile (configFile);
		if (savePlayers (global, env, configFile)) {
			fclose (configFile);
			perror ("savePlayers");
			exit (1);
		}
		fclose (configFile);
	} else {
		perror ("Saving Config");
		exit (1);
	}
	
	allegro_exit ();

	cout<<"We're currently recruiting C/C++ developers and a web designer"<<endl;
	cout<<"See http://atanks.sourceforge.net for the latest news and downloads"<<endl;
	
	return 0;
}


END_OF_MAIN ();
