/*
 * atanks - obliterate each other with oversize weapons
 * Copyright (C) 2003  Thomas Hudson
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * */


#include "gensound.h"

#define	SOUNDTYPE signed short int

double sin1DPoint (double amplitude, double scale, double xo, double lambda, int octaves)
{
	double	maxH = 0;
	double	h = 0;
	int iteration;

	for (iteration = 1; iteration <= octaves; iteration++) {
		double	zoom = scale / (iteration * iteration);

		h += amplitude * sin (xo / zoom);
		maxH += amplitude;

		amplitude *= lambda;
	}

	// Normalised
	return (h / maxH);
}

void GENSOUND::genThunder ()
{
	SOUNDTYPE *soundBuf = (SOUNDTYPE*)(sample.data);
	double intensity;
	int timeIndex = 0;

	double bang = 0;
	double noise = 0;
	double attackVal = 0;
	double decayVal = 1;
	long int bangCount = rand () % 30000;

	while (attackVal < decayVal && timeIndex < SOUNDLEN) {
	 	intensity = (perlin1DPoint (1.0, 16000, timeIndex + offset, 0.5, 6) + 1) / 2;
	 	attackVal += 0.005;
		soundBuf[timeIndex] += (int)(perlin1DPoint (1.0, 200, timeIndex + offset, 0.5, 6) * ((1 << (sizeof (SOUNDTYPE) * 8))) * attackVal * intensity);
		//soundBuf[timeIndex] += (int)(sin1DPoint (1.0, 100, timeIndex + offset, 0.5, 3) * ((1 << (sizeof (SOUNDTYPE) * 8)) / 2) * attackVal);
		timeIndex++;
	}
	while (decayVal > 0.01 && timeIndex < SOUNDLEN) {
	 	intensity = (perlin1DPoint (1.0, 16000, timeIndex + offset, 0.5, 6) + 1) / 2;
		if (!bangCount) {
			bang = 1.0;
			decayVal = MIN (decayVal * 1.5, 0.99);
			bangCount = rand () % 100000;
		}
		if (noise) {
			noise *= 0.99995;
			if (noise < 0.01)
				noise = 0;
		}
		if (bang) {
			if (intensity < 0) {
				intensity = MIN(-bang,intensity);
			} else {
				intensity = MAX(bang,intensity);
			}
			if (!noise) {
				if (!(rand () % 1000))
					noise = 1.0;
			}
			bang *= 0.9999;
			if (bang < 0.01)
				bang = 0;
		}
		soundBuf[timeIndex] += (int)(perlin1DPoint (1.0, 200, timeIndex + offset, 0.5 + noise * 0.45, 6) * (((1 << (sizeof (SOUNDTYPE) * 8)) / 2 - 1)) * decayVal * intensity);
		//soundBuf[timeIndex] += (int)(sin1DPoint (1.0, 100, timeIndex + offset, 0.5, 3) * ((1 << (sizeof (SOUNDTYPE) * 8)) / 2) * decayVal);
		decayVal *= 0.99997;
		timeIndex++;
		bangCount--;
	}
}

int listenerX, listenerY;
void processPoint (SOUNDTYPE *soundBuf, int x, int y)
{
	double intensity = fabs (perlin2DPoint (1.0, 100, x, y, 0.3, 6));
	int xdist = listenerX - x;
	int ydist = listenerY - y;
	double distance = sqrt ((xdist * xdist) + (ydist * ydist));
	int timeIndex = distance * 200;
	double attackVal = 0;
	double decayVal = 1 / (distance / 199) * intensity;
	long int offset;
	offset = rand ();

	while (attackVal < decayVal && timeIndex < SOUNDLEN) {
		attackVal += 0.01;
		//soundBuf[timeIndex] += (int)(perlin1DPoint (1.0, 100, timeIndex + offset, 0.2, 6) * ((1 << (sizeof (SOUNDTYPE) * 8)) / 3) * attackVal);
		soundBuf[timeIndex] += (int)(sin1DPoint (1.0, 100, timeIndex + offset, 0.5, 3) * ((1 << (sizeof (SOUNDTYPE) * 8)) / 3) * attackVal);
		timeIndex++;
	}
	while (decayVal > 0.01 && timeIndex < SOUNDLEN) {
		//soundBuf[timeIndex] += (int)(perlin1DPoint (1.0, 100, timeIndex + offset, 0.2, 6) * ((1 << (sizeof (SOUNDTYPE) * 8)) / 3) * decayVal);
		soundBuf[timeIndex] += (int)(sin1DPoint (1.0, 100, timeIndex + offset, 0.5, 3) * ((1 << (sizeof (SOUNDTYPE) * 8)) / 3) * decayVal);
		decayVal *= 0.99;
		timeIndex++;
	}
}

void processLineSound (SOUNDTYPE *soundBuf, int x1, int y1, int x2, int y2)
{
	int width = abs (x2 - x1);
	int height = abs (y2 - y1);
	int length = sqrt ((width * width) + (height * height));
	int pos;

	for (pos = 0; pos < length; pos++) {
		double where = (double)pos / length;
		int x = (int)(MIN(x1,x2) + (where * width));
		int y = (int)(MIN(y1,y2) + (where * height));
		processPoint (soundBuf, x, y);
	}
}

int main (int argc, char *argv[])
{
	long int timeIndex;
	SOUNDTYPE soundBuf[SOUNDLEN];
	int numPoints = 10;
	int points[20] = {100, 10,
			600, 10,
			200, 70,
			400, 80,
			500, 200,
			80, 300,
			500, 100,
			900, 300,
			200, 200,
			10, 100};
	int x1, x2, y1, y2;
	FILE *outfile;
	int point;

	listenerX = 100;
	listenerY = 100;
	srand (time ());
	offset = rand ();

	for (timeIndex = 0; timeIndex < SOUNDLEN; timeIndex++)
		soundBuf[timeIndex] = 0;

	genThunder (soundBuf);
	/*for (point = 1; point < numPoints; point++) {

		x1 = points[(point - 1) * 2];
		y1 = points[(point - 1) * 2 + 1];
		x2 = points[point * 2];
		y2 = points[point * 2 + 1];

		processLineSound (soundBuf, x1, y1, x2, y2);
	}*/

	outfile = fopen ("lightning.raw", "w");
	fwrite (soundBuf, sizeof (SOUNDTYPE), SOUNDLEN, outfile);
	fclose (outfile);
}
