/*
 * atanks - obliterate each other with oversize weapons
 * Copyright (C) 2003  Thomas Hudson
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * */


#include "environment.h"
#include "globaldata.h"
#include "explosion.h"
#include "missile.h"
#include "decor.h"
#include "tank.h"

MISSILE::MISSILE ()
{
	_align = LEFT;
	_current.x = 0;
	_current.y = 0;
	_current.w = 0;
	_current.h = 0;
}

MISSILE::~MISSILE ()
{
	_env->removeObject (this);
}

MISSILE::MISSILE (GLOBALDATA *global, ENVIRONMENT *env, double xpos, double ypos, double xvel, double yvel, int weaponType)
{
	initialise ();
	setEnvironment (env);
	player = NULL;
	_align = LEFT;
	_global = global;
	x = xpos;
	y = ypos;
	xv = xvel;
	yv = yvel;
	type = weaponType;
	if (type < WEAPONS)
		weap = &weapon[type];
	else
		weap = &naturals[type - WEAPONS];
	mass = weap->mass;
	drag = weap->drag;
	sound = weap->sound;
	expSize = weap->radius;
	etime = weap->etime;
	damage = weap->damage;
	eframes = weap->eframes;
	picpoint = weap->picpoint;
	noimpact = weap->noimpact;
	_bitmap = (BITMAP *)_global->gfxData.MI[picpoint].dat;
}

void MISSILE::initialise ()
{
	PHYSICAL_OBJECT::initialise ();
	s = 0;
	age = 0;
	if (type >= SML_METEOR && type <= LRG_METEOR) {
		angle = rand () % 360;
		spin = rand () % 20 - 10;
	} else {
		angle = 0;
		spin = 0;
	}
}

void MISSILE::explode ()
{
	EXPLOSION *explosion;

	explosion = new EXPLOSION (_global, _env, x, y, type);
	if (!explosion) {
		perror ("missile.cc: Failed allocating memory for explosion in MISSILE::explode");
		exit (1);
	}
	explosion->player = player;
	destroy = TRUE;
}

int MISSILE::triggerTest ()
{
	int quell = 1;
	TANK *tank;

	// Has it hit a tank?
	for (int objCount = 0; (tank = (TANK*)_env->getNextOfClass (TANK_CLASS, &objCount)) && tank; objCount++) {
		if (x > tank->x - TANKWIDTH && x < tank->x + TANKWIDTH && y > tank->y && y < tank->y + TANKHEIGHT && tank->l > 0) {
			hitSomething = 1;
			tank->requireUpdate ();
		}
	}
	if ((y > 0) && (noimpact || hitSomething || (y >= SCREEN_HEIGHT) || (getpixel (_env->terrain, (int)x, (int)y) != PINK))) {
		quell = 0;
		if (type >= SML_ROLLER && type <= DTH_ROLLER && !s) {
			if (age > 1) {
				quell = 1;
				if (x < 1)
					x = 1;
				if (x > (SCREEN_WIDTH - 2))
					x = (SCREEN_WIDTH - 2);
				s = 1;
			} else {
				s = 4;
			}
		} else if (type == CLUSTER || type == SUP_CLUSTER) {
			quell = 1;
			_env->am = 6;
			_env->realm = 5;
			x -= xv;
			y -= yv;
			for (int spreadCount = 0; spreadCount < 5; spreadCount++) {
				MISSILE *newmis;
				int newType;

				int ca = (int) (180 + ((SPREAD * spreadCount) - (SPREAD * 1.5)));

				if (type == CLUSTER)
					newType = BOMBLET;
				else
					newType = SUP_BOMBLET;

				newmis = new MISSILE (_global, _env, x, y - 20, _global->slope[ca][0] * 5 * (100.0 / FRAMES_PER_SECOND), _global->slope[ca][1] * 5 * (100.0 / FRAMES_PER_SECOND), newType);
				if (!newmis) {
					perror ("missile.cc: Failed allocating memory for newmis in MISSILE::triggerTest (CLUSTER)");
					exit (1);
				}
				newmis->s = 0;
				newmis->player = player;
				newmis->setUpdateArea
					((int) newmis->x - 4,
					 (int) newmis->y - 4, 8, 8);
			}
			destroy = TRUE;
		} else if (type == FUNKY_BOMB || type == FUNKY_DEATH) {
			quell = 1;
			_env->am = 10;
			_env->realm = 9;
			if (y > SCREEN_HEIGHT - 5)
				y = SCREEN_HEIGHT - 5;
			for (int spreadCount = 0; spreadCount < 9; spreadCount++) {
				MISSILE *newmis;
				int newType;

				int heading,range;
				if (type == FUNKY_BOMB) {
					range = rand() % 250;
					newType = FUNKY_BOMBLET;
				} else {
					range = rand() % 350;
					newType = FUNKY_DEATHLET;
				}

				heading = rand () % 180;
				while ((heading > 90) && (y + (_global->slope[heading][1] * range * 2) > SCREEN_HEIGHT))
					heading--;
				if (rand() % 2)
					heading = 360 - heading;

				newmis = new MISSILE (_global, _env, x, y, _global->slope[heading][0] / 2 * (100.0 / FRAMES_PER_SECOND), _global->slope[heading][1] / 2 * (100.0 / FRAMES_PER_SECOND), newType);
				if (!newmis) {
					perror ("missile.cc: Failed allocating memory for newmis in MISSILE::triggerTest (FUNKY)");
					exit (1);
				}
				newmis->s = 1;
				newmis->noimpact = TRUE;
				newmis->hitSomething = 0;
				newmis->countdown = range;
				newmis->player = player;
				newmis->setUpdateArea
					((int) newmis->x - 4,
					(int) newmis->y - 4, 8, 8);
			}
			destroy = TRUE;
		} else if (type == FUNKY_BOMBLET || type == FUNKY_DEATHLET) {
			quell = 1;
			if (countdown <= 0)
				quell = 0;
			countdown--;
			setUpdateArea ((int) x - 4, (int) y - 4, 8, 8);
		} else if (type >= SML_NAPALM && type <= LRG_NAPALM) {
			int numBlobs;
			quell = 1;

			if (type == SML_NAPALM)
				numBlobs = 10;
			else if (type == MED_NAPALM)
				numBlobs = 25;
			else
				numBlobs = 60;
			_env->am = numBlobs;
			_env->realm = numBlobs - 1;
			if (y > SCREEN_HEIGHT - 5)
				y = SCREEN_HEIGHT - 5;
			for (int spreadCount = 0; spreadCount < numBlobs; spreadCount++) {
				MISSILE *newmis;

				int heading,iVel;

				heading = 135 + (rand () % 90);
				iVel = rand () % (expSize / 15);

				newmis = new MISSILE (_global, _env, x, y - 20,
					_global->slope[heading][0] * iVel * (100.0 / FRAMES_PER_SECOND),
					_global->slope[heading][1] * iVel * (100.0 / FRAMES_PER_SECOND),
					NAPALM_JELLY);
				if (!newmis) {
					perror ("missile.cc: Failed allocating memory for newmis in MISSILE::triggerTest (NAPALM)");
					exit (1);
				}
				newmis->hitSomething = 0;
				newmis->player = player;
				newmis->setUpdateArea
					((int) newmis->x - 4,
					(int) newmis->y - 4, 8, 8);
			}
			destroy = TRUE;
		} else if ((type == BURROWER || type == PENETRATOR)) {
			if (!hitSomething)
				quell = 1;
			if (!s) {
				s = 1;
				xv *= 0.1;
				yv *= 0.1;
			}
		} else if (type >= RIOT_CHARGE && type <= RIOT_BLAST) {
			quell = 0;
			destroy = TRUE;
			setUpdateArea ((int) x - 4, (int) y - 4, 8, 8);
		}
		if (!quell) {
			_env->realm--;
			trigger ();
		}
	}

	return (!quell);
}

int MISSILE::applyPhysics ()
{
	TANK *ltank;
	int detonate = 0;
	age++;

	if (age > 15 * FRAMES_PER_SECOND || y < -65535)
		detonate = 1;
	if (!s) {
		if (type >= SML_METEOR && type <= LRG_METEOR) {
			angle += spin;
			angle = angle % 360;
		} else {
			angle = (int) (atan (yv / xv) * (180 / PI) * ACHANGE) - 64 + ((xv < 0) ? 128 : 0);
		}
		for (int objCount = 0; (ltank = (TANK*)_env->getNextOfClass (TANK_CLASS, &objCount)) && ltank; objCount++) {
			if (ltank->player != player) {
				double xaccel = 0, yaccel = 0;
				ltank->repulse (x, y, &xaccel, &yaccel);
				xv += xaccel;
				yv += yaccel;
			}
		}
		PHYSICAL_OBJECT::applyPhysics ();
		if (rand () % 5 == 0) {
			DECOR *decor;
			decor = new DECOR (_global, _env, x, y, xv, yv, expSize / 20, DECOR_SMOKE);
			if (!decor) {
				perror ("missile.cc: Failed allocating memory for decor in MISSILE::applyPhysics (METEOR)");
				exit (1);
			}
		}
	} else {
		if (type >= SML_ROLLER && type <= DTH_ROLLER) {
			if (s == 1) {
				for (y = SCREEN_HEIGHT - 1; getpixel (_env->terrain, (int)x, (int)y) != PINK; y--);
				if (getpixel (_env->terrain, (int)x + 1, (int)y + 1) == PINK)
					s = 3;
				else
					s = 2;
			}
			if (s == 2)
				x--, angle -= 3;
			if (s == 3)
				x++, angle += 3;
			if ((y > 0) && ((s == 4) || (x < 1) || (x > (SCREEN_WIDTH-1)) || (y > SCREEN_HEIGHT - 1) || (getpixel (_env->terrain, (int)x, (int)y) != PINK))) {
				detonate = 1;

			} else {
				// roll roller
				y += 3;
				if ((y > 0) && ((y > SCREEN_HEIGHT - 1) || (getpixel (_env->terrain, (int)x, (int)y) != PINK)))
					for (y = SCREEN_HEIGHT - 4; getpixel (_env->terrain, (int)x, (int)y + 3) != PINK; y--);
				if (y > (SCREEN_HEIGHT - 5) && !detonate)
					y = (SCREEN_HEIGHT - 5);
				if (angle < 0)
					angle += 256;
				if (angle > 255)
					angle -= 256;
			}
		} else if (type == FUNKY_BOMBLET || type == FUNKY_DEATHLET) {
			//double accel = (_env->wind - xv) / mass * drag;
			if (x + xv < 1 || x + xv > (SCREEN_WIDTH-1)) {
				xv = -xv;	//bounce on the border
			} else {
				x += xv;
			}
			if (y + yv >= SCREEN_HEIGHT)
			{
				yv = -yv * 0.5;
				xv *= 0.95;
			}
			y += yv;
		} else if (type == BURROWER || type == PENETRATOR) {
			angle = (int) (atan (yv / xv) * (180 / PI) * ACHANGE) - 64 + ((xv < 0) ? 128 : 0);
			for (int objCount = 0; (ltank = (TANK*)_env->getNextOfClass (TANK_CLASS, &objCount)) && ltank; objCount++) {
				if (ltank->player != player) {
					double xaccel = 0, yaccel = 0;
					ltank->repulse (x, y, &xaccel, &yaccel);
					xv += xaccel * 0.1;
					yv += yaccel * 0.1;
				}
			}
			if (x + xv < 1 || x + xv > (SCREEN_WIDTH-1)) {
				xv = -xv;	//bounce on the border
			}
			if (y + yv >= SCREEN_HEIGHT)
			{
				yv = -yv * 0.5;
				xv *= 0.95;
			}
			y += yv;
			x += xv;
			yv -= _env->gravity * 0.05 * (100.0 / FRAMES_PER_SECOND);
			if (getpixel (_env->terrain, (int)x, (int)y) == PINK) {
				detonate = TRUE;
			}
		}
	}
	if (detonate)
		hitSomething = 1;

	return (detonate);
}


void MISSILE::trigger ()
{
	EXPLOSION *explosion;

	if (type >= TREMOR && type <= TECTONIC)
		explosion = new EXPLOSION (_global, _env, x, y, xv, yv, type);
	else if (type >= RIOT_CHARGE && type <= RIOT_BLAST)
		explosion = new EXPLOSION (_global, _env, x, y, xv, yv, type);
	else
		explosion = new EXPLOSION (_global, _env, x, y, type);

	if (!explosion) {
		perror ("missile.cc: Failed allocating memory for explosion in MISSILE::trigger");
		exit (1);
	}
	explosion->player = player;
	destroy = TRUE;
	
	if ((type >= DIRT_BALL && type <= SUP_DIRT_BALL) || (type >= RIOT_CHARGE && type <= RIOT_BLAST)) {
		play_sample ((SAMPLE *) _global->SOUND[9].dat, 128 + (expSize / 2), 128, 1000 - (expSize * 2), 0);
	} else if (type == NAPALM_JELLY) {
		//play_sample ((SAMPLE *) _global->SOUND[9].dat, 128 + (expSize / 2), 128, 1000 - (expSize * 2), 0);
	} else {
		play_sample ((SAMPLE *) _global->SOUND[WEAPONSOUNDS + sound].dat, 255, 128, 1000, 0);
	}
}

void MISSILE::draw (BITMAP *dest)
{
	if (!destroy) {
		if (y < MENU) {
			BITMAP *bbitmap = _bitmap;
			double by = y;
			int bangle = angle;

			_bitmap = (BITMAP*)_global->gfxData.M[3].dat;
			y = (double)MENU + (_bitmap->h / 2);
			angle = 0;

			VIRTUAL_OBJECT::draw (dest);
			_bitmap = bbitmap;
			y = by;
			angle = bangle;
		} else {

			if (type == NAPALM_JELLY) {
				circlefill (_env->db, (int)x, (int)y, expSize, makecol16 (255, 0, 0));
				setUpdateArea ((int)x - (expSize + 1), (int)y - (expSize + 1), (expSize + 1) * 2, (expSize + 1) * 2);
				requireUpdate ();
			} else {
				VIRTUAL_OBJECT::draw (dest);
			}
		}
	}
}

int MISSILE::isSubClass (int classNum)
{
	if (classNum == MISSILE_CLASS)
		return (TRUE);
	else
		return (FALSE);
		//return (PHYSICAL_OBJECT::isSubClass (classNum));
}
