/*
 * atanks - obliterate each other with oversize weapons
 * Copyright (C) 2003  Thomas Hudson
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * */


#include "environment.h"
#include "globaldata.h"
#include "player.h"
#include "tank.h"
#include "menu.h"

// When defined draws AI 'planning'
//#define AI_PLANNING_DEBUG	1

PLAYER::PLAYER (GLOBALDATA *global, ENVIRONMENT *env)
{
	_global = global;
	_env = env;
	money = 15000;
	score = 0;
	played = 0;
	won = 0;
	tank = NULL;
	_turnStage = 0;
	selected = FALSE;

	nm[0] = 99;
	for (int count = 1; count < WEAPONS; count++)
		nm[count] = 0;

	for (int count = 0; count < ITEMS; count++)
		ni[count] = 0;

	strncpy (_name, "New", 11);
	type = HUMAN_PLAYER;

	generatePreferences ();
	vengeful = rand () % 100;
	vengeanceThreshold = (double)(rand () % 1000) / 1000.0;
	revenge = NULL;

	color = rand () % WHITE;
	typeText[0] = "Human";
	typeText[1] = "Guesser";
	typeText[2] = "Range Finder";
	typeText[3] = "Targetter";
	initMenuDesc ();
}

PLAYER::PLAYER (GLOBALDATA *global, ENVIRONMENT *env, FILE *file)
{
	_global = global;
	_env = env;
	money = 15000;
	score = 0;
	tank = NULL;
	_turnStage = 0;
	selected = FALSE;

	nm[0] = 99;
	for (int count = 1; count < WEAPONS; count++)
		nm[count] = 0;

	for (int count = 0; count < ITEMS; count++)
		ni[count] = 0;

	loadFromFile (file);
	type = (int)type;

	typeText[0] = "Human";
	typeText[1] = "Guesser";
	typeText[2] = "Range Finder";
	typeText[3] = "Targetter";
	initMenuDesc ();
}

void PLAYER::setComputerValues ()
{
	switch ((int)type) {
		case HUMAN_PLAYER:
			break;
		case GUESSER_PLAYER:
			rangeFindAttempts = 1;
			retargetAttempts = 0;
			focusRate = 0.6;
			errorMultiplier = 1.0;
			break;
		case RANGEFINDER_PLAYER:
			rangeFindAttempts = 10;
			retargetAttempts = 2;
			focusRate = 0.1;
			errorMultiplier = 0.3;
			break;
		case TARGETTER_PLAYER:
			rangeFindAttempts = 20;
			retargetAttempts = 5;
			focusRate = 1.0;
			errorMultiplier = 0.1;
			break;
		default:
			rangeFindAttempts = 0;
			retargetAttempts = 0;
			focusRate = 0;
			errorMultiplier = 1.0;
			break;
	}
}

int displayPlayerName (GLOBALDATA *global, ENVIRONMENT *env, int x, int y, void *data);
void PLAYER::initMenuDesc ()
{
	int destroyPlayer (GLOBALDATA *global, ENVIRONMENT *env, void *data);
	menudesc = new MENUDESC;
	if (!menudesc) {
		perror ("player.cc: Failed allocating memory for menudesc in PLAYER::initMenuDesc");
		exit (1);
	}
	menudesc->title = _name;

	// Name,Color
	menudesc->numEntries = 6;
	menudesc->okayButton = TRUE;
	menudesc->quitButton = FALSE;

	menuopts = new MENUENTRY[menudesc->numEntries];
	if (!menuopts) {
		perror ("player.cc: Failed allocating memory for menuopts in PLAYER::initMenuDesc");
		exit (1);
	}

	// Player name
	menuopts[0].name = "Name";
	menuopts[0].displayFunc = NULL;
	menuopts[0].color = color;
	menuopts[0].value = (double*)_name;
	menuopts[0].specialOpts = NULL;
	menuopts[0].type = OPTION_TEXTTYPE;
	menuopts[0].viewonly = FALSE;
	menuopts[0].x = HALF_WIDTH - 3;
	menuopts[0].y = HALF_HEIGHT - 68;

	// Player colour
	menuopts[1].name = "Colour";
	menuopts[1].displayFunc = NULL;
	menuopts[1].color = WHITE;
	menuopts[1].value = (double*)&color;
	menuopts[1].specialOpts = NULL;
	menuopts[1].type = OPTION_COLORTYPE;
	menuopts[1].viewonly = FALSE;
	menuopts[1].x = HALF_WIDTH - 3;
	menuopts[1].y = HALF_HEIGHT - 48;

	// Player type (human, computer)
	menuopts[2].name = "Type";
	menuopts[2].displayFunc = NULL;
	menuopts[2].color = WHITE;
	menuopts[2].value = (double*)&type;
	menuopts[2].min = 0;
	menuopts[2].max = LAST_PLAYER_TYPE - 1;
	menuopts[2].increment = 1;
	menuopts[2].defaultv = 0;
	menuopts[2].format = "%s";
	menuopts[2].specialOpts = typeText;
	menuopts[2].type = OPTION_SPECIALTYPE;
	menuopts[2].viewonly = FALSE;
	menuopts[2].x = HALF_WIDTH - 3;
	menuopts[2].y = HALF_HEIGHT - 28;

	menuopts[3].name = "Played";
	menuopts[3].displayFunc = NULL;
	menuopts[3].color = WHITE;
	menuopts[3].value = (double*)&played;
	menuopts[3].format = "%.0f";
	menuopts[3].specialOpts = NULL;
	menuopts[3].type = OPTION_DOUBLETYPE;
	menuopts[3].viewonly = TRUE;
	menuopts[3].x = HALF_WIDTH - 3;
	menuopts[3].y = HALF_HEIGHT - 8;

	menuopts[4].name = "Won";
	menuopts[4].displayFunc = NULL;
	menuopts[4].color = WHITE;
	menuopts[4].value = (double*)&won;
	menuopts[4].format = "%.0f";
	menuopts[4].specialOpts = NULL;
	menuopts[4].type = OPTION_DOUBLETYPE;
	menuopts[4].viewonly = TRUE;
	menuopts[4].x = HALF_WIDTH - 3;
	menuopts[4].y = HALF_HEIGHT + 12;

	menuopts[5].name = "Delete This Player";
	menuopts[5].displayFunc = NULL;
	menuopts[5].color = WHITE;
	menuopts[5].value = (double*)destroyPlayer;
	menuopts[5].data = (void*)this;
	menuopts[5].type = OPTION_ACTIONTYPE;
	menuopts[5].viewonly = FALSE;
	menuopts[5].x = HALF_WIDTH - 3;
	menuopts[5].y = HALF_HEIGHT + 100;

	menudesc->entries = menuopts;
}

PLAYER::~PLAYER ()
{
	if (tank)
		delete tank;

	delete menuopts;
	delete menudesc;
}

// Following are config file definitions
#define	GFILE_LATEST_VERSION	5

#define GFILE_NAME		_name, sizeof (char), 11, file
#define	GFILE_WEAPONPREFERENCE1	_weaponPreference, sizeof (int), 51, file
#define	GFILE_WEAPONPREFERENCE2	_weaponPreference, sizeof (int), 54, file
#define	GFILE_WEAPONPREFERENCE3	_weaponPreference, sizeof (int), 58, file
#define	GFILE_WEAPONPREFERENCE4	_weaponPreference, sizeof (int), 60, file
#define	GFILE_TYPE		&type, sizeof (double), 1, file
#define	GFILE_VENGEFUL		&vengeful, sizeof (int), 1, file
#define	GFILE_VENGEANCETHRESHOLD &vengeanceThreshold, sizeof (double), 1, file
#define	GFILE_COLOR		&color, sizeof (int), 1, file
#define	GFILE_COLOR2		&color2, sizeof (int), 1, file
#define	GFILE_PLAYED		&played, sizeof (double), 1, file
#define	GFILE_WON		&won, sizeof (double), 1, file

#define	GFILE_VERSION_1(a)	a (GFILE_NAME);\
				a (GFILE_TYPE);\
				a (GFILE_COLOR);\
				a (GFILE_WEAPONPREFERENCE1);\
				a (GFILE_PLAYED);\
				a (GFILE_WON);

#define	GFILE_VERSION_2(a)	a (GFILE_NAME);\
				a (GFILE_TYPE);\
				a (GFILE_COLOR);\
				a (GFILE_COLOR2);\
				a (GFILE_WEAPONPREFERENCE2);\
				a (GFILE_PLAYED);\
				a (GFILE_WON);

#define	GFILE_VERSION_3(a)	a (GFILE_NAME);\
				a (GFILE_TYPE);\
				a (GFILE_COLOR);\
				a (GFILE_COLOR2);\
				a (GFILE_WEAPONPREFERENCE3);\
				a (GFILE_PLAYED);\
				a (GFILE_WON);

#define	GFILE_VERSION_4(a)	a (GFILE_NAME);\
				a (GFILE_TYPE);\
				a (GFILE_COLOR);\
				a (GFILE_COLOR2);\
				a (GFILE_WEAPONPREFERENCE4);\
				a (GFILE_PLAYED);\
				a (GFILE_WON);

#define	GFILE_VERSION_5(a)	a (GFILE_NAME);\
				a (GFILE_VENGEFUL);\
				a (GFILE_VENGEANCETHRESHOLD);\
				a (GFILE_TYPE);\
				a (GFILE_COLOR);\
				a (GFILE_COLOR2);\
				a (GFILE_WEAPONPREFERENCE4);\
				a (GFILE_PLAYED);\
				a (GFILE_WON);

#define	GFILE_WRITE_LATEST	GFILE_VERSION_5(fwrite)
#define	GFILE_READ_VERSION_1	GFILE_VERSION_1(fread)
#define	GFILE_READ_VERSION_2	GFILE_VERSION_2(fread)
#define	GFILE_READ_VERSION_3	GFILE_VERSION_3(fread)
#define	GFILE_READ_VERSION_4	GFILE_VERSION_4(fread)
#define	GFILE_READ_VERSION_5	GFILE_VERSION_5(fread)

int PLAYER::saveToFile (FILE *file)
{
	int version = GFILE_LATEST_VERSION;
	if (!file)
		return (-1);

	fwrite (&version, sizeof (int), 1, file);

	GFILE_WRITE_LATEST

	return (ferror (file));
}

int PLAYER::loadFromFile (FILE *file)
{
	int version;
	if (!file)
		return (-1);

	fread (&version, sizeof (int), 1, file);

	switch (version) {
		case 1:
			GFILE_READ_VERSION_1;
			break;
		case 2:
			GFILE_READ_VERSION_2;
			break;
		case 3:
			GFILE_READ_VERSION_3;
			break;
		case 4:
			GFILE_READ_VERSION_4;
			break;
		case 5:
			GFILE_READ_VERSION_5;
			break;
		default:
			break;
	}
	if (GFILE_LATEST_VERSION > version)
		generatePreferences ();

	return (ferror (file));
}

void PLAYER::newRound ()
{
	setComputerValues ();
	if (!tank) {
		tank = new TANK(_global, _env);
		if (!tank) {
			perror ("player.cc: Failed allocating memory for tank in PLAYER::newRound");
			exit (1);
		}
		tank->player = this;
	}
	tank->newRound ();
}

void PLAYER::initialise ()
{
	nm[0] = 99;
	for (int count = 1; count < WEAPONS; count++)
		nm[count] = 0;

	for (int count = 0; count < ITEMS; count++)
		ni[count] = 0;
}

void PLAYER::generatePreferences ()
{
	char accounted[THINGS+1];
	int thing;

	memset (accounted, 0, THINGS + 1);
	for (int count = 0; count < THINGS + 1; count++) {
		thing = rand () % (THINGS + 1);
		while (accounted[thing]) {
			thing = rand () % (THINGS + 1);
		}
		accounted[thing] = 1;
		_weaponPreference[count] = thing;
	}
}

void PLAYER::setName (char *name)
{
	strncpy (_name, name, 10);
}

int PLAYER::controlTank (TANK *ctank)
{
	if (key[KEY_F1])
		save_bmp ("scrnshot.bmp", _env->db, NULL);

	if (key_shifts & KB_CTRL_FLAG && ctrlUsedUp) {
		if (key[KEY_LEFT] || key[KEY_RIGHT] ||
			key[KEY_UP] || key[KEY_DOWN])
			ctrlUsedUp = TRUE;
		else
			ctrlUsedUp = FALSE;
	} else {
		ctrlUsedUp = FALSE;
	}

	k = 0;
	if (keypressed () && !fi) {
		k = readkey ();

		if ((_env->stage == 3) && (k >> 8 == KEY_ENTER || k >> 8 == KEY_ESC || k >> 8 == KEY_SPACE))
			return (-1);
		if (k >> 8 == KEY_ESC) {
			void clockadd ();
			install_int_ex (clockadd, SECS_TO_TIMER(6000));
			int mm = _env->ingamemenu ();
			install_int_ex (clockadd, BPS_TO_TIMER (100));
			_env->make_update (0, 0, SCREEN_WIDTH, SCREEN_HEIGHT);
			_env->make_bgupdate (0, 0, SCREEN_WIDTH, SCREEN_HEIGHT);
			if (mm == 1) {
				_global->command = 1;
				return (-1);
			}
			if (mm == 2) {
				_global->command = 0;
				return (-1);
			}
		}
	}
	if ((int)type == HUMAN_PLAYER || !ctank) {
		return (humanControls (ctank));
	} else if (!_env->stage) {
		return (computerControls (ctank));
	}
	return (0);
}

int PLAYER::chooseItemToBuy ()
{
	for (int count = 0; count < THINGS; count++) {
		int currItem = _weaponPreference[count];
		int nextMod = ((currItem < WEAPONS)?nm[currItem]:ni[currItem - WEAPONS]) / 5;
		if (!_env->isItemAvailable (currItem))
			continue;
		if (rand () % (2 + nextMod))
			continue;
		if (currItem < WEAPONS) {
			if (currItem == 0) {
				continue;
			}
			if (money >= weapon[currItem].cost)
				return (currItem);
		} else {
			if (money >= item[currItem - WEAPONS].cost)
				return (currItem);
		}
	}
	return (-1);
}

char *PLAYER::selectRevengePhrase (double scale)
{
	static char *insults[] = {
		"Grr!",
		"I worship the ground that awaits your corpse",
		"You are not even beneath my contempt",
		"Your father was a hamster!",
		"Your mother smells of elderberries!",
		"Curse you!",
		"I'll get you next time!",
		"I'll get you!",
		"YOU RAT!",
		"Hey, mush-brain!",
		"It's only a flesh-wound!",
		"Je ne suis pas un poisson",	// I am not a fish
		"I hate you!",
		"I'll be back!",
		"You'll suffer for this!",
		"You'll regret that!",
	};

	return (insults[rand () % 15]);
}

char *PLAYER::selectGloatPhrase (double scale)
{
	return ("");
}

int PLAYER::traceShellTrajectory (TANK *ctank, int *estimateX, int *estimateY, double targetX, double targetY, double x, double y, double xv, double yv)
{
	TANK *ltank;
	int hitSomething = 0;
	double distance;
	char direction = (char)copysign (1, xv);
	double drag;
	int ticks = 0;
	int maxTicks = 500;

	drag = weapon[ctank->cw].drag;
	if (ni[ITEM_DIMPLEP]) {
		drag *= item[ITEM_DIMPLEP].vals[0];
	} else if (ni[ITEM_SLICKP]) {
		drag *= item[ITEM_SLICKP].vals[0];
	}

	while (((getpixel (_env->terrain, (int)x, (int)y) == PINK) ||
		y < MENU || ticks > maxTicks) &&
		!hitSomething) {
		for (int objCount = 0; (ltank = (TANK*)_env->getNextOfClass (TANK_CLASS, &objCount)) && ltank; objCount++) {
			if (ltank->player != this) {
				double xaccel = 0, yaccel = 0;
				ltank->repulse (x, y, &xaccel, &yaccel);
				xv += xaccel;
				yv += yaccel;
			}
		}

		if (x + xv < 1 || x + xv > (SCREEN_WIDTH-1)) {
			xv = -xv;	//bounce on the border
			direction = -direction;
		} else {
			// motion - wind affected
			double accel = (_env->wind - xv) / weapon[ctank->cw].mass * drag * _env->viscosity;
			xv += accel;
			x += xv;
		}
		if (y + yv >= SCREEN_HEIGHT)
		{
			yv = -yv * 0.5;
			xv *= 0.95;
			if (fabs(xv) + fabs(yv) < 0.8)
				hitSomething = 1;
		}
		y += yv;
		yv += _env->gravity * (100.0 / FRAMES_PER_SECOND);
#ifdef	AI_PLANNING_DEBUG
		// Plot fire-paths for debugging purposes
		circlefill (screen, (int)x, (int)y, 2, color);
#endif
		ticks++;
	}
#ifdef	AI_PLANNING_DEBUG
	// Targetting circle for debugging purposes
	circlefill (screen, (int)targetX, (int)targetY, 20, color);
#endif

	*estimateX = (int)x;
	*estimateY = (int)y;
	distance = vector_length_f ((float)(x - targetX), (float)(y - targetY), 0.0);
	if (ticks > maxTicks)
		_targetPower /= 2;
	if ((targetX - x < 0) && (direction < 0))
		distance = -distance;
	else if ((targetX - x > 0) && (direction > 0))
		distance = -distance;
	return ((int)distance);
}

int PLAYER::rangeFind (TANK *ctank, int *estimateX, int *estimateY, int maxAttempts)
{
	int findCount;
	int spread = weapon[ctank->cw].spread;
	int overshoot = 1;

	findCount = 0;
	while (overshoot && (findCount < maxAttempts)) {
		spread = 1;
		for (z = 0; z < spread; z++) {
			double mxv,myv;
			int ca;

			ca = _targetAngle + ((SPREAD * z) - (SPREAD * (spread - 1) / 2));
			mxv = _global->slope[ca][0] * _targetPower / 100;
			myv = _global->slope[ca][1] * _targetPower / 100;

			overshoot = (int)(traceShellTrajectory (
				ctank,
				estimateX, estimateY,
				_target->x, _target->y,
				ctank->x + (_global->slope[ca][0] * GUNLENGTH) - mxv,
				ctank->y + (_global->slope[ca][1] * GUNLENGTH) - myv,
					mxv, myv));
			if (overshoot < 0) {
				_targetPower += (int)((rand () % (abs (overshoot) + 10)) * focusRate);
			} else {
				_targetPower -= (int)((rand () % (abs (overshoot) + 10)) * focusRate);
			}
			if (abs (overshoot) < abs (ctank->smallestOvershoot)) {
				ctank->smallestOvershoot = overshoot;
				ctank->bestPower = _targetPower;
				ctank->bestAngle = _targetAngle;
			}
			findCount++;
		}
	}
	return (overshoot);
}

int PLAYER::calculateDirectAngle (int dx, int dy)
{
	double angle;

	angle = atan2 ((double)dx, (double)dy) / PI * 180;
	angle += (rand () % 40 - 20) * errorMultiplier;

	if (angle < 0)
		angle = angle + 360;
	if (angle < 90)
		angle = 135;
	else if (angle > 270)
		angle = 225;

	return ((int)angle);
}

int PLAYER::calculateAngle (int dx, int dy)
{
	double angle;

	angle = atan2 ((double)dx, (double)dy - abs (dx)) / PI * 180;
	angle += (rand () % 40 - 20) * errorMultiplier;

	if (angle < 0)
		angle = angle + 360;
	if (angle < 90)
		angle = 135;
	else if (angle > 270)
		angle = 225;

	return ((int)angle);
}

int PLAYER::calculatePowerGivenAngle (int dx, int dy, int angle)
{
	double airTime, dxTime;
	double requiredPower;
	double slopeX, slopeY;
	double powerMult;
	double minYv, minV;

	slopeX = _global->slope[angle][1];
	slopeY = _global->slope[angle][0];

	/*
		The greater dx, the longer the missile will be in the air.
		Steeper angle, longer air time.
		The greater dy, the shorter air time.
		Double the gravity, double the power.
		Following wind = less power needed.
	*/


	//maxHeight = 0.5 * (yv * yv) / _env->gravity * (100.0 / FRAMES_PER_SECOND);
	// calculate minimum V to reach height dy
	if (dy > 0)
		minYv = sqrt (dy * _env->gravity * 2 * (100.0 / FRAMES_PER_SECOND));
	else
		minYv = 0;
	if ((slopeY != 0) && (minYv != 0))
		minV = fabs (minYv / slopeY);
	else
		minV = 0;
	if (slopeX != 0) {
		dxTime = (dx/fabs(slopeX));
	} else {
		// entirely down to the elements now
		dxTime = (dx / 0.000001);
	}
	//windAccel = (_env->wind / 20 * 0.2 * _env->viscosity);
	//dxTime = dxTime - (0.5 * windAccel * (dxTime*dxTime));
	if (dx < 0) {
		if (dxTime > 0)
			return (-1); // Won't ever get there
	} else {
		if (dxTime < 0)
			return (-1); // Won't ever get there
	}
	// Low target, less power
	// xdistance proportional to sqrt(dy)
	airTime = fabs(dxTime) + ((dy * slopeY) * _env->gravity * (100.0 / FRAMES_PER_SECOND)) * 2;

	// Less airTime doesn't necessarily mean less power
	// Horizontal firing means more power needed even though
	//   airTime is minimised
	powerMult = 1;
//	powerMult = (slopeX * slopeY);
//	if (powerMult == 0)
//		powerMult = 0.01;
	requiredPower = (sqrt (airTime * _env->gravity * (100.0 / FRAMES_PER_SECOND)) / powerMult) * 100;
	requiredPower += (int)((rand () % 100 - 50) * errorMultiplier);
	return ((int)requiredPower);
}

int PLAYER::adjustAngleGivenClearance (TANK *ctank, int xdist, int ydist, int clearance)
{
	while (!ctank->shootClearance (_targetAngle, clearance)) {
		if (xdist > 0) {
			if (_targetAngle < 270) {
				_targetAngle++;
			} else {
				_targetAngle += (int)((double)(rand () % 40 - 20) / (rangeFindAttempts + 1));
				break;
			}
		} else {
			if (_targetAngle > 90) {
				_targetAngle--;
			} else {
				_targetAngle += (int)((double)(rand () % 40 - 20) / (rangeFindAttempts + 1));
				break;
			}
		}
	}
	if (_targetAngle > 270)
		_targetAngle = 270 - (_targetAngle - 270);
	else if (_targetAngle < 90)
		_targetAngle = 90 + (90 - _targetAngle);

	return (_targetAngle);
}

TANK *PLAYER::selectTarget ()
{
	int objCount = 0;
	int tankNum = rand () % (_global->numTanks - 1);
	TANK *ltank = NULL;

	if (revenge && vengeful > rand () % 100)
			ltank = revenge->tank;
	if (!ltank) {
		while (((ltank = (TANK*)_env->getNextOfClass (TANK_CLASS, &objCount)) && (tankNum > 0)) || (ltank && (ltank->player == this))) {
			objCount++;
			tankNum--;
		}
	}
	return (ltank);
}

int PLAYER::computerControls (TANK *ctank)
{
	// At the most basic: select target, select weapon, aim and fire
	ctank->requireUpdate ();
	if (_turnStage == SELECT_TARGET) {
		_target = selectTarget ();
		if (!_target)
			return (-1);
		_turnStage++;
	} else if (_turnStage == SELECT_WEAPON) {
		int xdist = (int)(_target->x - ctank->x);
		int ydist = (int)(_target->y - ctank->y);
		int sdistance = (int)sqrt ((double)(xdist * xdist) + (double)(ydist * ydist));

		ctank->cw = 0;
		if (ctank->howBuried () > 135) {
			if (nm[RIOT_BLAST] > 0)
				ctank->cw = RIOT_BLAST;
			else if (nm[RIOT_CHARGE] > 0)
				ctank->cw = RIOT_CHARGE;
			else if (nm[RIOT_BOMB] > 0)
				ctank->cw = RIOT_BOMB;
			else if (nm[HVY_RIOT_BOMB] > 0)
				ctank->cw = HVY_RIOT_BOMB;
			else if (ni[ITEM_TELEPORT] > 0)
				ctank->cw = WEAPONS + ITEM_TELEPORT;
			else if (nm[SML_LAZER] > 0)
				ctank->cw = SML_LAZER;
			else if (nm[MED_LAZER] > 0)
				ctank->cw = MED_LAZER;
			else if (nm[LRG_LAZER] > 0)
				ctank->cw = LRG_LAZER;
		}
		if (ctank->cw == 0) {
			for (int weapCount = 0; weapCount < THINGS; weapCount++) {
				if (rand () % 2)
					continue;
				ctank->cw = _weaponPreference[weapCount];

				if (ctank->cw < WEAPONS) {
					if (!nm[ctank->cw])
						continue;
					if (sdistance < weapon[ctank->cw].radius * 1.5) {
						if (!(ctank->cw >= RIOT_CHARGE && ctank->cw <= RIOT_CHARGE))
							continue;
					} else if (ctank->cw >= RIOT_CHARGE && ctank->cw <= RIOT_CHARGE) {
						continue;
					} else if (ctank->cw >= SML_LAZER && ctank->cw <= LRG_LAZER && ydist > 0) {
						continue;
					}
				} else {
					int itemNum = ctank->cw - WEAPONS;
					if (!item[itemNum].selectable)
						continue;
					if (itemNum == ITEM_TELEPORT) {
						if (!ni[ITEM_PARACHUTE])
							continue;
					}
					if ((itemNum >= ITEM_VENGEANCE) &&
						(itemNum <= ITEM_FATAL_FURY)) {
						if ((ctank->l + ctank->sh < 20)
							&& (rand () % 100 == 0))
							break;
					}
				}
				break;
			}
		}
		if ((!nm[ctank->cw]) || (ctank->cw >= THINGS))
			ctank->cw = 0;
		_global->updateMenu = 1;
		_turnStage++;
	} else if (_turnStage == CALCULATE_ATTACK) {
		int overshoot;
		int estimateX, estimateY;
		int xdist = (int)(_target->x - ctank->x);
		int ydist = (int)(_target->y - ctank->y);
		//int sdistance = (int)sqrt ((double)(xdist * xdist) + (double)(ydist * ydist));

		ctank->smallestOvershoot = SCREEN_WIDTH;
		if (ctank->howBuried () > 135) {
			_targetAngle = 180;
			_targetPower = MAX_POWER / 10;
		} else {
			if (ctank->cw >= SML_LAZER && ctank->cw <= LRG_LAZER) {
				_targetAngle = calculateDirectAngle (xdist, ydist);
			} else {
				_targetAngle = calculateAngle (xdist, ydist);
				_targetAngle = adjustAngleGivenClearance (ctank, xdist, ydist, 60);
				_targetPower = calculatePowerGivenAngle (xdist, ydist, _targetAngle);
				if (rangeFindAttempts > 0)
					overshoot = rangeFind (ctank, &estimateX, &estimateY, rangeFindAttempts);
				else
					overshoot = 0;
				if (abs (overshoot) > weapon[ctank->cw].radius) {
					for (int retargetCount = 0; (retargetCount < retargetAttempts) && (abs (overshoot) > weapon[ctank->cw].radius); retargetCount++) {
						if (rand () % 20 < 10) {
							_targetAngle = (rand () % 180) + 90;
						} else {
							_target = selectTarget ();
							if (!_target)
								return (-1);
							_targetAngle = calculateAngle (xdist, ydist);
							_targetAngle = adjustAngleGivenClearance (ctank, xdist, ydist, 60);
							_targetPower = calculatePowerGivenAngle (xdist, ydist, _targetAngle);
						}
						if (rangeFindAttempts > 0)
							overshoot = rangeFind (ctank, &estimateX, &estimateY, rangeFindAttempts);
						else
							overshoot = 0;
					}
				}
				if (abs (ctank->smallestOvershoot) > weapon[ctank->cw].radius * 3) {
					_targetAngle = calculateAngle (xdist, ydist);
					_targetAngle = adjustAngleGivenClearance (ctank, xdist, ydist, (int)(weapon[ctank->cw].radius * 1.5));
					_targetPower = (MAX_POWER / 2);
				} else {
					_targetAngle = ctank->bestAngle;
					_targetPower = ctank->bestPower;
				}
			}
		}
		_turnStage++;
	} else if (_turnStage == AIM_WEAPON) {
		if (_targetAngle > ctank->a && ctank->a < 270) {
			// Left
			ctank->a++;
			_global->updateMenu = 1;
		} else if (_targetAngle < ctank->a && ctank->a > 90) {
			// Right
			ctank->a--;
			_global->updateMenu = 1;
		} else if (_targetPower < (ctank->p - 3) && ctank->p > 0) {
			// Reduce power
			ctank->p -= 5;
			_global->updateMenu = 1;
		} else if (_targetPower > (ctank->p + 3) && ctank->p < MAX_POWER) {
			// Increase Power
			ctank->p += 5;
			_global->updateMenu = 1;
		} else {
			_turnStage++;
		}
	} else if (fi) {
		// if (fi) don't do any of the following
	} else if (_turnStage == FIRE_WEAPON) {
		ctank->activateCurrentSelection ();
		_turnStage = SELECT_TARGET;
	}
	return (0);
}

int PLAYER::humanControls (TANK *ctank)
{
	if (ctank) {
		if (!_env->mouseclock && mouse_b & 1 && mouse_x >= 250
		    && mouse_x < 378 && mouse_y >= 11 && mouse_y < 19) {
			_global->updateMenu = 1;
			if (ctank->fs) {
				ctank->sht++;
			}
			ctank->fs = 1;
			if (ctank->sht > SHIELDS - 1) {
				ctank->sht = 0;
			}
		}
		if (!_env->mouseclock && mouse_b & 1 && mouse_x >= 250
		    && mouse_x < 378 && mouse_y >= 21
		    && mouse_y < 29 && ctank->player->ni[ctank->sht] > 0 && (ctank->fs || ctank->sh > 0)) {
			_global->updateMenu = 1;
			ctank->ds = ctank->sht;
			ctank->player->ni[ctank->sht]--;
			ctank->sh = (int)item[ctank->sht].vals[SHIELD_ENERGY];
		}
	
		ctank->requireUpdate ();
	}
	
	//Keyboard control
	if (!_env->stage) {
		if (ctank) {
			if (key[KEY_LEFT] && !ctrlUsedUp && ctank->a < 270) {
				ctank->a++;
				_global->updateMenu = 1;
				if (key_shifts & KB_CTRL_FLAG)
					ctrlUsedUp = TRUE;
			}
			if (key[KEY_RIGHT] && !ctrlUsedUp && ctank->a > 90) {
				ctank->a--;
				_global->updateMenu = 1;
				if (key_shifts & KB_CTRL_FLAG)
					ctrlUsedUp = TRUE;
			}
			if (key[KEY_DOWN] && !ctrlUsedUp && ctank->p > 0) {
				ctank->p -= 5;
				_global->updateMenu = 1;
				if (key_shifts & KB_CTRL_FLAG)
					ctrlUsedUp = TRUE;
			}
			if (key[KEY_UP] && !ctrlUsedUp && ctank->p < MAX_POWER) {
				ctank->p += 5;
				_global->updateMenu = 1;
				if (key_shifts & KB_CTRL_FLAG)
					ctrlUsedUp = TRUE;
			}
		}
	}
	if (k && !fi) {
		if (!_env->stage) {
			if (ctank) {
				if (k >> 8 == KEY_N) {
					ctank->a = 180;
					_global->updateMenu = 1;
				}
				if (k >> 8 == KEY_TAB) {
					_global->updateMenu = 1;
					while (1) {
						ctank->cw++;
						if (ctank->cw >= THINGS)
							ctank->cw = 0;
						if (ctank->cw < WEAPONS) {
							if (ctank->player->nm[ctank->cw])
								break;
						} else {
							if (item[ctank->cw - WEAPONS].selectable && ctank->player->ni[ctank->cw - WEAPONS])
								break;

						}
					}
				}
				if ((k >> 8 == KEY_SPACE) &&
					(((ctank->cw < WEAPONS) && (ctank->player->nm[ctank->cw] > 0)) ||
						((ctank->cw < THINGS) && (ctank->player->ni[ctank->cw - WEAPONS] > 0)))) {
					ctank->activateCurrentSelection ();
				}
			}
		}
		if ((_env->stage == 3) && (k >> 8 == KEY_ENTER || k >> 8 == KEY_ESC || k >> 8 == KEY_SPACE))
			return (-1);
	}
	return (0);
}
