/*
 * atanks - obliterate each other with oversize weapons
 * Copyright (C) 2003  Thomas Hudson
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * */


#include "main.h"
#include "menu.h"

enum playerType {
	HUMAN_PLAYER,
	GUESSER_PLAYER,
	RANGEFINDER_PLAYER,
	TARGETTER_PLAYER,
	LAST_PLAYER_TYPE,
};
enum turnStages {
	SELECT_TARGET,
	SELECT_WEAPON,
	CALCULATE_ATTACK,
	AIM_WEAPON,
	FIRE_WEAPON,
};

class TANK;
class PLAYER {
	private:
	char	_name[11];
	int	_currTank;
	GLOBALDATA *_global;
	ENVIRONMENT *_env;
	char	_turnStage;
	TANK	*_target;
	int	_targetAngle;
	int	_targetPower;
	int	_overshoot;
	int	_weaponPreference[THINGS];

	public:
	double	type;
	PLAYER	*revenge;
	int	vengeful;	// 0-100 chance of retaliation
	double	vengeanceThreshold;	// Damage required to warrant revenge
					//   (% of max armour)
	double	annoyanceFactor;
	int	rangeFindAttempts;
	int	retargetAttempts;
	double	focusRate;
	double	errorMultiplier;
	int	score;
	double	played, won;
	double	selected;
	int	money;
	TANK	*tank;
	int	color;
	int	color2;
	int 	nm[WEAPONS], ni[ITEMS];
	MENUENTRY	*menuopts;
	MENUDESC	*menudesc;
	char *typeText[LAST_PLAYER_TYPE];
	

	PLAYER (GLOBALDATA *global, ENVIRONMENT *env);
	PLAYER (GLOBALDATA *global, ENVIRONMENT *env, FILE *file);
	~PLAYER	();
	void	setName (char *name);
	char	*getName () {return (_name);};
	void	newRound ();
	void	initialise ();
	TANK	*nextTank ();
	TANK	*currTank ();
	int	controlTank (TANK *ctank);
	int	humanControls (TANK *ctank);
	int	computerControls (TANK *ctank);
	int	traceShellTrajectory (TANK *ctank, int *estimateX, int *estimateY, double targetX, double targetY, double x, double y, double xv, double yv);
	int	rangeFind (TANK *ctank, int *estimateX, int *estimateY, int maxAttempts);
	int	calculatePowerGivenAngle (int dx, int dy, int angle);
	int	calculateAngle (int dx, int dy);
	int	calculateDirectAngle (int dx, int dy);
	int	adjustAngleGivenClearance (TANK *ctank, int xdist, int ydist, int clearance);
	TANK	*selectTarget ();
	int	chooseItemToBuy ();
	void	generatePreferences ();
	void	setComputerValues ();
	char	*selectRevengePhrase (double scale);
	char	*selectGloatPhrase (double scale);

	int	saveToFile (FILE *file);
	int	loadFromFile (FILE *file);
	void	initMenuDesc ();
};
